# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This is for advanced machine learning models or complex data structures and algorithms, presented in a digestible interface
"""
from typing import Callable, Union, List, overload, Iterator, Any, Set, Tuple
from k1lib.cli.init import BaseCli, fastF
import k1lib.cli as cli; import k1lib, os, math
from k1lib.cli.typehint import *
import numpy as np; from collections import deque
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["embed", "complete", "kmeans", "tsne", "bloom"]
settings = k1lib.Settings().add("cuda", None, "whether to run the models on the GPU or not. True for GPU, False for CPU. None (default) for GPU if available, else CPU")
k1lib.settings.cli.add("models", settings, "settings related to k1lib.cli.models");
_cuda = k1lib.Wrapper(None)
def cuda() -> bool: # internal func to figure out whether the funcs should run on gpus or not # cuda
    if _cuda() is None: _cuda.value = torch.cuda.is_available() and torch.cuda.device_count() >= 1 # cuda
    return _cuda()                                                               # cuda
sentence_transformers = k1lib.dep("sentence_transformers"); embed_models_cpu = dict(); embed_models_cuda = dict() # cuda
def embed_models(): return embed_models_cuda if cuda() else embed_models_cpu     # embed_models
settings.add("embed", k1lib.Settings().add("model", "all-MiniLM-L6-v2", "what model to choose from `SentenceTransformer` library").add("bs", 512, "batch size to feed the model. For all-MiniLM-L6-v2, it seems to be able to deal with anything. I've tried 10k batch and it's still doing good")) # embed_models
def embed_model(): # returns correct function capable of passing in List[str] and will spit out np.ndarray with shape (N, F) # embed_model
    modelName = settings.embed.model                                             # embed_model
    if modelName not in embed_models():                                          # embed_model
        model = sentence_transformers.SentenceTransformer(modelName)             # embed_model
        if cuda(): model = model.cuda()                                          # embed_model
        embed_models()[modelName] = model.encode                                 # embed_model
    return embed_models()[modelName]                                             # embed_model
class embed(BaseCli):                                                            # embed
    def __init__(self, norm=True):                                               # embed
        """Gets an embedding vector for every sentence piped into this.
Example::

    # returns (384,)
    "abc" | embed() | shape()
    # returns (2, 384). Don't worry that this is less performant, as behind the scenes, it will automatically batch all lines together and pass through the model only once
    ["abc", "def"] | embed().all() | shape()

There are several settings you can set::

    settings.cli.models.embed.model = "msmarco-distilbert-base-v4" # specifies model used in this function
    settings.cli.models.cuda = False # tells the system to only use the CPU to run the models

I'd suggest picking "all-MiniLM-L6-v2" for general purpose tasks, and "msmarco-distilbert-base-v4"
for document-lookup style applications.

:param norm: whether to normalize the output embeddings or not"""                # embed
        self.model = embed_model(); self.normF = (lambda x: (x - (x | cli.toMean())) / (x | cli.toStd())) if norm else (lambda x: x) # embed
    def __ror__(self, it): return self.normF(self.model([it])[0] if isinstance(it, str) else self.model(list(it))) # embed
    def _all_opt(self, it:List[str]): return it | cli.batched(settings.embed.bs, True) | cli.apply(self.__ror__) | cli.joinStreams() # embed
transformers = k1lib.dep("transformers"); generic_models_cpu = dict(); generic_models_cuda = dict() # embed
def generic_models(): return generic_models_cuda if cuda() else generic_models_cpu # generic_models
settings.add("generic", k1lib.Settings().add("model", "google/flan-t5-xl", "what model to choose from `transformers` library").add("bs", 16, "batch size to feed the model. For flan-t5-xl, 16 seems to be the sweet spot for 24GB VRAM (RTX 3090/4090). Decrease it if you don't have as much VRAM")) # generic_models
def generic_model(maxTokens=100): # returns correct function capable of passing in str|List[str] and will spit out List[str] # generic_model
    modelName = settings.generic.model                                           # generic_model
    if modelName not in generic_models():                                        # generic_model
        if modelName.startswith("google/flan-t5-"):                              # generic_model
            tokenizer = transformers.T5Tokenizer.from_pretrained(modelName, **({"device_map": "auto"} if cuda() else {})) # generic_model
            model = transformers.T5ForConditionalGeneration.from_pretrained(modelName, **({"device_map": "auto"} if cuda() else {})) # generic_model
            conf = transformers.GenerationConfig(max_new_tokens=maxTokens); cuda_ = cuda() # generic_model
            # if cuda_: model = model.cuda()                                     # generic_model
            def inner(it):                                                       # generic_model
                inputs = tokenizer(list(it), return_tensors="pt", padding=True).input_ids # generic_model
                if cuda_: inputs = inputs.cuda()                                 # generic_model
                return (tokenizer.decode(line) for line in model.generate(inputs, conf)) # generic_model
            generic_models()[modelName] = inner                                  # generic_model
        else: raise Exception("Currently, only Google T5 Flan models are supported") # generic_model
    return generic_models()[modelName]                                           # generic_model
class complete(BaseCli):                                                         # complete
    def __init__(self, prompt:str=""):                                           # complete
        """Uses a LLM to autocomplete something.
Example::

    # returns "4". In case you're living in a cage, these LLMs are not entirely math savants. They sure understand English though
    "What is 2 + 6?" | complete()
    # returns ["4", "4"]
    ["What is 2 + 6?", "What is 8 + 2?"] | complete().all() | deref()

Can change model type by doing ``settings.cli.models.generic.model = "google/flan-t5-xl"``

:param prompt: if specified, will inject this bit of text after all of the inputs.
    Can be something like "Please translate the above paragraph to German"
"""                                                                              # complete
        self.model = generic_model(); self.prompt = prompt                       # complete
    def __ror__(self, it):                                                       # complete
        arrMode = not isinstance(it, str); prompt = self.prompt                  # complete
        it = (list(it) if arrMode else [it]) | cli.apply(lambda x: f"{x}\n\n\n{prompt}: ") | cli.deref() # complete
        ans = self.model(it) | cli.apply(lambda x: x.replace("<pad>", "").replace("<unk>", "").replace("</s>", "").strip()) # complete
        return ans if arrMode else ans | cli.item()                              # complete
    def _all_opt(self, it:List[str]):                                            # complete
        return it | cli.batched(settings.generic.bs, True) | cli.apply(self.__ror__) | cli.joinStreams() # complete
skclus = k1lib.dep("sklearn.cluster")                                            # complete
skpre = k1lib.dep("sklearn.preprocessing")                                       # complete
skmet = k1lib.dep("sklearn.metrics")                                             # complete
def refine(fea, a, b, kwargs, timeout=1):                                        # refine
    scores = []; topScore = float("-inf")                                        # refine
    for k in torch.loglinspace(a, b, 10).numpy().astype(int) | cli.aS(np.unique): # refine
        km = skclus.KMeans(n_clusters=k, **{**{"init": "k-means++", "n_init": 10, "max_iter": 30}, **kwargs}) # refine
        try: [fea] | cli.applyTh(km.fit, timeout=timeout) | cli.ignore()         # refine
        except: break                                                            # refine
        score = skmet.silhouette_score(fea, km.labels_)                          # refine
        topScore = max(topScore, score); scores.append([k, score])               # refine
        if score*2 < topScore: break                                             # refine
    return scores | ~cli.sort(1) | cli.cut(0) | cli.item(), k                    # refine
def findCenters(fea, c, kwargs, timeout=1):                                      # findCenters
    if c is None:                                                                # findCenters
        a = 2; b = len(fea); c, b = refine(fea, a, b, kwargs, timeout); a = round(a*0.2 + c*0.8); b = round(b*0.5 + c*0.5) # findCenters
        while b - a > 3:     c, b = refine(fea, a, b, kwargs, timeout); a = round(a*0.2 + c*0.8); b = round(b*0.5 + c*0.5) # findCenters
    km = skclus.KMeans(n_clusters=c, init="k-means++", n_init=10, max_iter=100)  # findCenters
    km.fit(fea); return km.cluster_centers_, km.labels_                          # findCenters
class kmeans(BaseCli):                                                           # kmeans
    def __init__(self, k=None, mode=1, timeout=1, **kwargs):                     # kmeans
        """Do k-means clustering, returning the cluster centers.
Example::

    features, true_labels = sklearn.datasets.make_blobs(n_samples=1_000, centers=5, cluster_std=0.5)
    centers = features | kmeans();
    centers | shape() # likely return (5, 2)

    # plotting things out
    plt.plot(*features.T, ".")
    plt.plot(*centers.T,  ".")

.. image:: ../images/kmeans.png

:param k: if specified, will use this k value. Else tries to guess what the best value is
:param mode: mode 0 (returns [cluster centers, labels]), mode 1 (returns cluster centers only), mode 2 (returns labels only)
:param timeout: internally will try kmeans for up to this number of seconds only.
    Will kill the job if it's taking longer
:param kwargs: keyword arguments will be passed into sklearn.cluster.KMeans directly.
    Some interesting parameters include ``init``, ``n_init``, ``max_iter``
"""                                                                              # kmeans
        self.k = k; self.mode = mode; self.timeout = timeout; self.kwargs = kwargs # kmeans
    def __ror__(self, it):                                                       # kmeans
        scaler = skpre.StandardScaler(); fea = scaler.fit_transform(it); mode = self.mode # kmeans
        centers, labels = findCenters(fea, self.k, self.kwargs, self.timeout)    # kmeans
        centers = scaler.inverse_transform(centers)                              # kmeans
        if mode == 0: return [centers, labels]                                   # kmeans
        return centers if mode == 1 else labels                                  # kmeans
skmani = k1lib.dep("sklearn.manifold")                                           # kmeans
class tsne(BaseCli):                                                             # tsne
    def __init__(self, n=2, **kwargs):                                           # tsne
        """Transforms feature vectors of shape (N, F) down to (N, 2) for easy plotting.
Example::

    from sklearn.datasets import make_blobs
    features, true_labels = make_blobs(n_samples=1_000, n_features=5, centers=5, cluster_std=0.2)

    features | shape()          # returns (1000, 5)
    features | tsne() | shape() # returns (1000, 2)
    # plotting things out that has nice colors and whatnot
    features | tsne() & kmeans(5, 2) | ~aS(lambda xy,c: plt.scatter(*xy.T,c=c))

.. image:: ../images/tsne.png

:param n: number of output components (aka size of feature vector)
:param kwargs: other keyword arguments passed into ``sklearn.manifold.TSNE``"""  # tsne
        self.n = n; self.kwargs = kwargs                                         # tsne
    def __ror__(self, it):                                                       # tsne
        if not isinstance(it, k1lib.settings.cli.arrayTypes): it = np.array(list(it)) # tsne
        return skmani.TSNE(self.n, **self.kwargs).fit_transform(it)              # tsne
settings.add("bloom", k1lib.Settings().add("scalable", k1lib.Settings().add("capacity", 1000, "initial filter's capacity").add("growth", 4, "how fast does the filter's capacity grow over time when the capacity is reached"), "settings for when you don't declare the bloom's capacity ahead of time"), "bloom filter settings") # tsne
pybloom_live = k1lib.dep("pybloom_live", "pybloom-live", "https://github.com/joseph-fox/python-bloomfilter, https://pypi.org/project/pybloom-live/") # tsne
class bloom(BaseCli):                                                            # bloom
    def __init__(self, n:int=None, p:float=0.1, overflow:bool=False):            # bloom
        """Creates a bloom filter.
Example::

    bf = ["raptor", "skylake", "merlin", "twinscan nxt", "sapphire rapids"] | bloom()
    "raptor"       in bf # returns True
    "twinscan nxt" in bf # returns True
    "twin"         in bf # most likely returns False, small chance returns True

This also allows distributed computing quite easily::

    bf = range(10) | applyMp(lambda i: cat(f"file-{i}.txt") | bloom()) | bloom.join()

This code assumes that you have 10 files filled with text with file names "file-1.txt" and
you want to check whether a string exists or not in all of those files. It is expected that

.. admonition:: Scalable bloom filter

    It's possible to leave the filter's capacity empty, which will create an initial filter
    with capacity 1000. When that capacity is reached, it will expand the filter to a capacity
    of 4000. Then 16000, and so on. This can be tweaked in the settings::

        settings.cli.models.bloom.scalable.capacity = 2000 # sets filter's default initial capacity
        settings.cli.models.bloom.scalable.growth = 2      # sets filter's growth factor when it runs out of space

    Because fundamentally, bloom filters can't grow, internally, this will create multiple bloom
    filters with increasing capacity, and whenever you search for a term, it will have to search
    through multiple filters to get the answer. So even though you can leave the capacity empty,
    it will degrade performance a little bit, which might be undesirable.

:param n: number of elements to be stored, aka filter's capacity
:param p: false positive probability (put 0.1 for 10% false positive, 0.01 for 1%)
:param overflow: if True, allows append more elements than the capacity of the filter, else (default) don't allow it""" # bloom
        self.n = n; self.p = p; self.overflow = overflow                         # bloom
    def __ror__(self, it):                                                       # bloom
        n = self.n; p = self.p                                                   # bloom
        bf = pybloom_live.ScalableBloomFilter(settings.bloom.scalable.capacity, p, settings.bloom.scalable.growth) if n is None else pybloom_live.BloomFilter(n, p) # bloom
        if self.overflow:                                                        # bloom
            for e in it:                                                         # bloom
                try: bf.add(e)                                                   # bloom
                except IndexError: bf.count = 0; bf.add(e)                       # bloom
        else:                                                                    # bloom
            for e in it: bf.add(e)                                               # bloom
        return bf                                                                # bloom
    @staticmethod                                                                # bloom
    def join():                                                                  # bloom
        def inner(bfs):                                                          # bloom
            bfs = list(bfs)                                                      # bloom
            if len(set([type(b) for b in bfs])) > 1: raise Exception("Can't join normal filters and scalable filters together. Please specify a common capacity to all of the bloom filters") # bloom
            if type(bfs[0]) == pybloom_live.BloomFilter: # my implementation, which should be a little faster than builtin .union() # bloom
                b = bfs[0].copy()                                                # bloom
                for b_ in bfs[1:]: b.bitarray |= b_.bitarray                     # bloom
                return b                                                         # bloom
            else: # has to use complex .union() in this case                     # bloom
                b = bfs[0]                                                       # bloom
                for b_ in bfs[1:]: b = b.union(b_)                               # bloom
                return b                                                         # bloom
        return cli.aS(inner)                                                     # bloom
try:                                                                             # bloom
    import pybloom_live                                                          # bloom
    @k1lib.patch(pybloom_live.pybloom.BloomFilter)                               # bloom
    def __or__(self, other):                                                     # bloom
        if isinstance(other, BaseCli): return other.__ror__(self)                # bloom
        return self.union(other)                                                 # bloom
except: pass                                                                     # bloom



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This is for functions that sort of changes the table
structure in a dramatic way. They're the core transformations
"""
from typing import List, Union, Iterator, Callable, Any, Tuple, Dict
from collections import defaultdict, Counter, deque
from k1lib.cli.init import patchDefaultDelim, BaseCli, oneToMany, fastF, yieldT
import k1lib.cli as cli; import k1lib.cli.init as init; from k1lib.cli.typehint import *
import itertools, numpy as np, k1lib, math, json; plt = k1lib.dep("matplotlib.pyplot")
try: import torch; hasTorch = True
except: torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = True
__all__ = ["transpose", "T", "reshape", "insert", "splitW", "splitC",
           "joinStreams", "joinSt", "joinStreamsRandom", "activeSamples",
           "table", "batched", "batchedTrigger", "window", "groupBy", "ungroup",
           "insertColumn", "insertIdColumn", "insId",
           "unsqueeze",
           "count", "hist", "permute", "AA_", "peek", "peekF",
           "repeat", "repeatF", "repeatFrom", "oneHot", "latch"]
settings = k1lib.settings.cli
class transpose(BaseCli):                                                        # transpose
    def __init__(self, dim1:int=0, dim2:int=1, fill=None):                       # transpose
        """Join multiple columns and loop through all rows. Aka transpose.
Example::

    # returns [[1, 4], [2, 5], [3, 6]]
    [[1, 2, 3], [4, 5, 6]] | transpose() | deref()
    # returns [[1, 4], [2, 5], [3, 6], [0, 7]]
    [[1, 2, 3], [4, 5, 6, 7]] | transpose(fill=0) | deref()

Multidimensional transpose works just like :meth:`torch.transpose` too::

    # returns (2, 7, 5, 3), but detected Tensor, so it will use builtin :meth:`torch.transpose`
    torch.randn(2, 3, 5, 7) | transpose(3, 1) | shape()
    # also returns (2, 7, 5, 3), but actually does every required computation. Can be slow if shape is huge
    torch.randn(2, 3, 5, 7) | deref(igT=False) | transpose(3, 1) | shape()

Can also work with numpy arrays::

    # returns (5, 3, 2)
    np.random.randn(2, 3, 5) | transpose(0, 2) | op().shape

Be careful with infinite streams, as transposing stream of shape (inf, 5) will
hang this operation! Either don't do it, or temporarily limit all infinite streams like
this::

    with settings.cli.context(inf=21):
        # returns (3, 21)
        [2, 1, 3] | repeat() | transpose() | shape()

Also be careful with empty streams, as you might not get any results at all::

    # returns [], as the last stream has no elements
    [[1, 2], [3, 4], []] | transpose() | deref()
    # returns [[1, 3, 0], [2, 4, 0]]
    [[1, 2], [3, 4], []] | transpose(fill=0) | deref()

This also has an alias of ``T()``, so you can make it shorter, like ``np.random.randn(2, 3, 5) | T()``

:param fill: if not None, then will try to zip longest with this fill value"""   # transpose
        super().__init__(); self.fill = fill                                     # transpose
        self.d1 = min(dim1, dim2); self.d2 = max(dim1, dim2)                     # transpose
        self.normal = self.d1 == 0 and self.d2 == 1                              # transpose
    def _all_array_opt(self, it, level:int): return it | transpose(self.d1 + level, self.d2 + level, self.fill) # transpose
    def _typehint(self, inp):                                                    # transpose
        if isinstance(inp, tArrayTypes):                                         # transpose
            if inp.rank is None: return inp                                      # transpose
            if inp.rank > max(self.d1, self.d2): return inp                      # transpose
            else: return tAny() # this case doesn't quite exist                  # transpose
        if self.d1 == 0 and self.d2 == 1:                                        # transpose
            if isinstance(inp, tListIterSet):                                    # transpose
                if isinstance(inp.child, tListIterSet):                          # transpose
                    return tIter(tList(inp.child.child))                         # transpose
        return tAny()                                                            # transpose
    def __ror__(self, it):                                                       # transpose
        d1 = self.d1; d2 = self.d2; fill = self.fill                             # transpose
        if isinstance(it, torch.Tensor): return it.transpose(d1, d2)             # transpose
        if isinstance(it, np.ndarray):                                           # transpose
            dims = list(range(len(it.shape)))                                    # transpose
            temp = dims[d1]; dims[d1] = dims[d2]; dims[d2] = temp                # transpose
            return it.transpose(dims)                                            # transpose
        if d1 != 0 or d2 != 1: return it | cli.serial(*([transpose(fill=fill).all(i) for i in range(d1, d2)] + [transpose(fill=fill).all(i-1) for i in range(d2-1, d1, -1)])) # transpose
        if self.fill is None: return zip(*it)                                    # transpose
        else: return itertools.zip_longest(*it, fillvalue=fill)                  # transpose
    @staticmethod                                                                # transpose
    def fill(fill="", dim1:int=0, dim2:int=1):                                   # transpose
        """Convenience method to fill in missing elements of a table.
Example::

    # returns [[1, 2, 3], [4, 5, 0]]
    [[1, 2, 3], [4, 5]] | transpose.fill(0) | deref()
    # also returns [[1, 2, 3], [4, 5, 0]], demonstrating how it works underneath
    [[1, 2, 3], [4, 5]] | transpose(fill=0) | transpose(fill=0) | deref()"""     # transpose
        return transpose(dim1, dim2, fill=fill) | transpose(dim1, dim2, fill=fill) # transpose
    @staticmethod                                                                # transpose
    def wrap(f, dim1:int=0, dim2:int=1, fill=None):                              # transpose
        """Wraps ``f`` around 2 :class:`transpose`, can be useful in combination with
:class:`k1lib.cli.init.mtmS`. Example::

    # returns [[1, 4, 3, 4], [8, 81, 10, 11]]
    [range(1, 5), range(8, 12)] | transpose.wrap(mtmS.f(apply(op()**2), 1)) | deref()
    # also returns [[1, 4, 3, 4], [8, 81, 10, 11]], demonstrating the typical way to do this
    [range(1, 5), range(8, 12)] | apply(op()**2, 1) | deref()

The example given is sort of to demonstrate this only. Most of the time, just use
:class:`~k1lib.cli.modifier.apply` with columns instead. But sometimes you need direct
access to a column, so this is how you can do it."""                             # transpose
        if not isinstance(f, BaseCli): f = cli.applyS(f)                         # transpose
        return transpose(dim1, dim2, fill) | f | transpose(dim1, dim2, fill)     # transpose
    def _jsF(self, meta):                                                        # transpose
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # transpose
        if self.d1 != 0 or self.d2 != 1: raise Exception(f"transpose._jsF() doesn't allow complex transpose across many dimensions yet") # transpose
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.transpose({cli.kjs.v(self.fill)})", fIdx # transpose
#tOpt.clearPasses()                                                              # transpose
def oTranspose1(cs, ts, metadata): # `transpose() | transpose().all()` to `transpose() | apply(aS(torch.stack) | transpose())` # oTranspose1
    tr, ap = cs; t = ts[0]                                                       # oTranspose1
    if (not ap.normal) or (not tr.normal): return None                           # oTranspose1
    if (not isinstance(ap.f, transpose)) or (not ap.f.normal): return None       # oTranspose1
    if isinstance(t, tListIterSet) and (isinstance(t.child, tTensor) or isinstance(t.child, tCollection) and isinstance(tLowest(*t.child.children), tTensor)): # oTranspose1
        return [transpose(), cli.apply(cli.aS(torch.stack) | transpose())]       # oTranspose1
tOpt.addPass(oTranspose1, [transpose, cli.apply])                                # oTranspose1
T = transpose                                                                    # oTranspose1
def _formStructure(it, dims, dimI):                                              # _formStructure
    if dimI >= len(dims): return next(it)                                        # _formStructure
    return [_formStructure(it, dims, dimI+1) for i in range(dims[dimI])]         # _formStructure
class reshape(BaseCli):                                                          # reshape
    def __init__(self, *dims):                                                   # reshape
        """Reshapes the input stream into the desired shape.
Example::

    # returns [[0, 1, 2], [3, 4, 5]]
    range(6) | reshape(2, 3) | deref()
    # returns [[0, 1], [2, 3], [4, 5]]
    range(6) | reshape(3, 2) | deref()
    # returns [[0, 1], [2, 3], [4, 5]], stopped early
    range(100) | reshape(3, 2) | deref()
    # returns [[0, 1, 2], [3, 4, 5]], can leave out first dimension
    range(6) | reshape(-1, 3) | deref()
    # returns [[0, 1, 2]], won't include 2nd element, as it ran out of elements
    range(5) | reshape(-1, 3) | deref()
    # throws error, as it ran out of elements and can't fulfill the request
    range(6) | reshape(3, 3) | deref()

Unlike :meth:`torch.reshape`, the input piped into this has to be a simple iterator.
If you have a complex data structure with multiple dimensions, turn that into a simple
iterator with :class:`joinStreams` first, like this::

    # returns [[[0, 1, 2]], [[3, 4, 5]]]
    [[[0], [1]], [[2], [3]], [[4], [5]]] | joinStreams(2) | reshape(2, 1, 3) | deref()""" # reshape
        self.dims = dims                                                         # reshape
    def __ror__(self, it):                                                       # reshape
        it = iter(it)                                                            # reshape
        if self.dims[0] == -1:                                                   # reshape
            try:                                                                 # reshape
                while True: yield _formStructure(it, self.dims, 1)               # reshape
            except StopIteration: pass                                           # reshape
        else:                                                                    # reshape
            for i in range(self.dims[0]): yield _formStructure(it, self.dims, 1) # reshape
class insert(BaseCli):                                                           # insert
    def __init__(self, element, begin=True):                                     # insert
        """Join element into list.
Example::

    # returns [5, 2, 6, 8]
    [2, 6, 8] | insert(5) | deref()
    # returns [2, 6, 8, 5]
    [2, 6, 8] | insert(5, begin=False) | deref()

    # returns [[3, 1], 2, 6, 8]
    [2, 6, 8] | insert([3, 1]) | deref()

:param element: the element to insert"""                                         # insert
        super().__init__(); self.element = element; self.begin = begin; self.expand = False # insert
    def __ror__(self, it):                                                       # insert
        element = self.element; it = iter(it)                                    # insert
        if self.begin: yield element; yield from it                              # insert
        else: yield from it; yield element                                       # insert
    def _jsF(self, meta):                                                        # insert
        fIdx = init._jsFAuto(); elemIdx = init._jsDAuto(); dataIdx = init._jsDAuto() # insert
        return f"const {elemIdx} = {json.dumps(self.element)}; const {fIdx} = ({dataIdx}) => {dataIdx}.insert({elemIdx}, {cli.kjs.v(self.begin)})", fIdx # insert
class splitW(BaseCli):                                                           # splitW
    def __init__(self, *weights:List[float]):                                    # splitW
        """Splits elements into multiple weighted lists. If no weights are provided,
then automatically defaults to [0.8, 0.2]. Example::

    # returns [[0, 1, 2, 3, 4, 5, 6, 7], [8, 9]]
    range(10) | splitW(0.8, 0.2) | deref()
    # same as the above
    range(10) | splitW() | deref()

This also works with array types::

    torch.randn(100, 3) | splitW() # returns 2 tensors with shapes (80, 3) and (20, 3)

See also: :class:`splitC`"""                                                     # splitW
        super().__init__();                                                      # splitW
        if len(weights) == 0: weights = [0.8, 0.2]                               # splitW
        self.weights = np.array(weights)                                         # splitW
    def __ror__(self, it):                                                       # splitW
        try: it[0]; len(it)                                                      # splitW
        except: it = list(it)                                                    # splitW
        ws = self.weights; c = 0; ws = (ws * len(it) / ws.sum()).astype(int)     # splitW
        for w in ws[:-1]: yield it[c:c+w]; c += w                                # splitW
        yield it[c:]                                                             # splitW
    def _jsF(self, meta):                                                        # splitW
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # splitW
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.splitW({cli.kjs.vs(self.weights) | cli.join(', ')})", fIdx # splitW
class splitC(BaseCli):                                                           # splitC
    def __init__(self, *checkpoints:List[float]):                                # splitC
        """Splits elements into multiple checkpoint-delimited lists.
Example::

    # returns [[0, 1], [2, 3, 4], [5, 6, 7, 8, 9]]
    range(10) | splitC(2, 5) | deref()
    # returns ['01', '234', '56789']
    "0123456789" | splitC(2, 5) | deref()

Here, you're specifying 2 checkpoints, 2 and 5, so it will split
the list up into 3 sections. First section is 0-2, second section
is 2-5, third section is 5-end. You can pass in fractional
checkpoints too::

    # returns [[0, 1], [2, 3, 4, 5], [6, 7, 8, 9]]
    range(10) | splitC(0.2, 0.6) | deref()

This cli might be unintuitive to remember, so if you want to just
split it up into 2 parts, check out :meth:`~k1lib.cli.filt.head.split`.

If you want to split things up by weighted length, check
out :class:`splitW`"""                                                           # splitC
        self.checkpoints = checkpoints | cli.aS(np.array)                        # splitC
        self.intMode = checkpoints | cli.apply(lambda x: int(x) == x) | cli.aS(all) # splitC
    def __ror__(self, it):                                                       # splitC
        try: it[0]; len(it)                                                      # splitC
        except: it = list(it)                                                    # splitC
        cs = self.checkpoints                                                    # splitC
        if not self.intMode: cs = (cs * len(it)).astype(int)                     # splitC
        cs = sorted(cs); yield it[:cs[0]]                                        # splitC
        for i in range(len(cs)-1): yield it[cs[i]:cs[i+1]]                       # splitC
        yield it[cs[-1]:]                                                        # splitC
    def _jsF(self, meta):                                                        # splitC
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # splitC
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.splitC({cli.kjs.vs(self.checkpoints) | cli.join(', ')})", fIdx # splitC
class joinStreams(BaseCli):                                                      # joinStreams
    def __init__(self, dims=1):                                                  # joinStreams
        """Joins multiple streams.
Example::

    # returns [1, 2, 3, 4, 5]
    [[1, 2, 3], [4, 5]] | joinStreams() | deref()
    # returns [[0, 1], [2], [3, 4, 5], [6, 7, 8], [], [9, 10]]
    [[[0, 1], [2], [3, 4, 5]], [[6, 7, 8], [], [9, 10]]] | joinStreams() | deref()
    # returns [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    [[[0, 1], [2], [3, 4, 5]], [[6, 7, 8], [], [9, 10]]] | joinStreams(2) | deref()

If you pass in :class:`numpy.ndarray` or :class:`torch.Tensor`, then it will
automatically use the C-accelerated version, like this::

    # returns Tensor with shape (6, 4)
    torch.randn(2, 3, 4) | joinStreams()
    # returns array with shape (6, 4)
    np.random.randn(2, 3, 4) | joinStreams()

Sometimes, you may want to impose some dimensional structure after joining all streams
together, which :class:`reshape` does.

If "joinStreams" is too unintuitive to remember, there's also an alias called
:class:`joinSt`.

:param dims: how many ``joinStreams()`` do you want to do consecutively?"""      # joinStreams
        if dims < 0: raise AttributeError(f"`dims` ({dims}) can't be less than 0, as it doesn't make any sense!") # joinStreams
        self.dims = dims; self.multi = cli.serial(*(joinStreams() for d in range(dims))) if dims > 1 else None # joinStreams
    def _all_array_opt(self, it, level:int): sh = it | cli.shape(); return it.reshape([*sh[:level], sh[level:level+self.dims+1] | cli.toProd(), *sh[level+self.dims+1:]]) # joinStreams
    def __ror__(self, streams:Iterator[Iterator[Any]]) -> Iterator[Any]:         # joinStreams
        if isinstance(streams, settings.arrayTypes): return self._all_array_opt(streams, 0) # joinStreams
        if self.multi != None:                                                   # joinStreams
            return streams | self.multi                                          # joinStreams
            def gen(): yield from streams | self.multi                           # joinStreams
            return gen()                                                         # joinStreams
        elif self.dims == 1:                                                     # joinStreams
            def gen():                                                           # joinStreams
                for stream in streams: yield from stream                         # joinStreams
            return gen()                                                         # joinStreams
        else: return streams                                                     # joinStreams
    def _jsF(self, meta):                                                        # joinStreams
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # joinStreams
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.joinSt({cli.kjs.v(self.dims)})", fIdx # joinStreams
joinSt = joinStreams                                                             # joinStreams
def probScale(ps, t): # t from 0 -> 1, for typical usage                         # probScale
    l = np.log(ps); avg = l.mean()                                               # probScale
    a = (l-avg)*t+avg; a -= a.max()                                              # probScale
    ans = np.exp(a); return ans/ans.sum()                                        # probScale
import random                                                                    # probScale
def rand(n, ps=None):                                                            # rand
    if ps is None:                                                               # rand
        while True: yield random.randrange(n)                                    # rand
    else:                                                                        # rand
        while True: yield from np.random.choice(n, size=100, p=ps)               # rand
class joinStreamsRandom(BaseCli):                                                # joinStreamsRandom
    def __init__(self, alpha=0, ps=None):                                        # joinStreamsRandom
        """Join multiple streams randomly. If any streams runs out, then quits. If
any stream yields :data:`~k1lib.cli.init.yieldT`, then just ignores that result and
continue. Could be useful in active learning. Example::

    # could return [0, 1, 10, 2, 11, 12, 13, ...], with max length 20, typical length 18
    [range(0, 10), range(10, 20)] | joinStreamsRandom() | deref()

    stream2 = [[-5, yieldT, -4, -3], yieldT | repeat()] | joinStreams()
    # could return [-5, -4, 0, -3, 1, 2, 3, 4, 5, 6], demonstrating yieldT
    [range(7), stream2] | joinStreamsRandom() | deref()

By default, all streams are treated equally, and are yielded with equal probabilities.
However, you can tweak these probabilities a little bit, to your liking. This is
controlled by the parameter ``alpha``:

.. image:: ../images/probScale.png

If ``alpha`` is 0, then all probabilities will be the same. If ``alpha`` is 1,
then all probabilities are proportional to the length of the input stream. The
original intention was to vary ``alpha`` just from 0 to 1, but it can actually
be of any number::

    [range(0, 10), range(10, 100)] | joinStreamsRandom(0) | shape(0) # returns around 21, because it favors both streams equally
    [range(0, 10), range(10, 100)] | joinStreamsRandom(1) | shape(0) # returns around 90, because it favors the second array 9x more
    [range(0, 10), range(10, 100)] | joinStreamsRandom(100) | shape(0) # returns 90, because it highly favors the second array
    [range(0, 10), range(10, 100)] | joinStreamsRandom(-100) | shape(0) # returns 10, because it highly favors the first array

:param alpha: if not zero, does a weighted joining, instead of totally uniform probability
:param ps: if specified, use these probabilities, else try to determine from the lengths of the input streams""" # joinStreamsRandom
        super().__init__(); self.alpha = alpha; self.ps = ps                     # joinStreamsRandom
    def __ror__(self, streams:Iterator[Iterator[Any]]) -> Iterator[Any]:         # joinStreamsRandom
        alpha = self.alpha; ps = self.ps; streams = list(streams); nStreams = len(streams) # joinStreamsRandom
        if alpha != 0:                                                           # joinStreamsRandom
            if ps is None:                                                       # joinStreamsRandom
                try: ps = np.array([len(st) for st in streams])                  # joinStreamsRandom
                except:                                                          # joinStreamsRandom
                    streams = [list(st) for st in streams]                       # joinStreamsRandom
                    ps = np.array([len(st) for st in streams])                   # joinStreamsRandom
            else: ps = np.array(list(ps))*1.0                                    # joinStreamsRandom
        else: ps = np.array([1/nStreams]*nStreams)                               # joinStreamsRandom
        ps = probScale(ps/ps.sum(), alpha)                                       # joinStreamsRandom
        streams = [iter(st) for st in streams]                                   # joinStreamsRandom
        try:                                                                     # joinStreamsRandom
            for streamIdx in rand(len(streams), ps):                             # joinStreamsRandom
                o = next(streams[streamIdx])                                     # joinStreamsRandom
                if not o is yieldT: yield o # "not is" to fix numpy `==`         # joinStreamsRandom
        except StopIteration: pass                                               # joinStreamsRandom
    def _jsF(self, meta):                                                        # joinStreamsRandom
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # joinStreamsRandom
        if self.alpha != 0 and not self.ps is None: raise Exception("joinStreamsRandom._jsF() doesn't allow custom alpha and ps yet") # joinStreamsRandom
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.joinStreamsRandom()", fIdx # joinStreamsRandom
class activeSamples(BaseCli):                                                    # activeSamples
    def __init__(self, limit:int=100, p:float=0.95):                             # activeSamples
        """Yields active learning samples.
Example::

    o = activeSamples()
    ds = range(10) # normal dataset
    ds = [o, ds] | joinStreamsRandom() # dataset with active learning capability
    next(ds) # returns 0
    next(ds) # returns 1
    next(ds) # returns 2
    o.append(20)
    next(ds) # can return     3     or 20
    next(ds) # can return (4 or 20) or 4

So the point of this is to be a generator of samples. You can define your dataset
as a mix of active learning samples and standard samples. Whenever there's a data
point that you want to focus on, you can add it to ``o`` and it will eventially yield
it.

.. warning::

    It might not be a good idea to set param ``limit`` to higher numbers than
    100. This is because, the network might still not understand a wrong sample
    after being shown multiple times, and will keep adding that wrong sample
    back in, distracting it from other samples, and reduce network's accuracy
    after removing active learning from it.

    If ``limit`` is low enough (from my testing, 30-100 should be fine), then
    old wrong samples will be kicked out, allowing for a fresh stream of wrong
    samples coming in, and preventing the problem above. If you found that
    removing active learning makes the accuracy drops dramatically, then try
    decreasing the limit.

:param limit: max number of active samples. Discards samples if number of samples
    is over this.
:param p: probability of actually adding the samples in"""                       # activeSamples
        super().__init__(); self.p = p                                           # activeSamples
        self.samples = deque([], limit)                                          # activeSamples
    def append(self, item):                                                      # activeSamples
        """Adds 1 sample."""                                                     # activeSamples
        if random.random() < self.p: self.samples.append(item)                   # activeSamples
    def extend(self, items):                                                     # activeSamples
        """Adds multiple samples."""                                             # activeSamples
        for item in items: self.append(item)                                     # activeSamples
    def __iter__(self):                                                          # activeSamples
        samples = self.samples                                                   # activeSamples
        while True:                                                              # activeSamples
            if len(samples) == 0: yield yieldT                                   # activeSamples
            else: yield samples.popleft()                                        # activeSamples
def table(delim:str=None):                                                       # table
    """Basically ``op().split(delim).all()``. This exists because this is used
quite a lot in bioinformatics. Example::

    # returns [['a', 'bd'], ['1', '2', '3']]
    ["a|bd", "1|2|3"] | table("|") | deref()"""                                  # table
    return cli.op().split(patchDefaultDelim(delim)).all()                        # table
def _batch(it, bs, includeLast, incl):                                           # _batch
    l = []; it = iter(it); sentinel = object(); last = sentinel                  # _batch
    try:                                                                         # _batch
        if incl:                                                                 # _batch
            while True:                                                          # _batch
                if last is not sentinel: l.append(last)                          # _batch
                for i in range(bs): l.append(next(it))                           # _batch
                last = next(it); l.append(last)                                  # _batch
                yield l; l = []                                                  # _batch
        else:                                                                    # _batch
            while True:                                                          # _batch
                for i in range(bs): l.append(next(it))                           # _batch
                yield l; l = []                                                  # _batch
            pass                                                                 # _batch
    except StopIteration:                                                        # _batch
        if includeLast and len(l) > 0: yield l                                   # _batch
def _batchRange(it, bs, includeLast, incl):                                      # _batchRange
    start, stop, step = it.start, it.stop, it.step                               # _batchRange
    lastCur = start; cur = lastCur + bs*step                                     # _batchRange
    if incl:                                                                     # _batchRange
        while cur <= stop: yield range(lastCur, min(cur+1, stop), step); lastCur = cur; cur += bs*step # _batchRange
    else:                                                                        # _batchRange
        while cur <= stop: yield range(lastCur, cur, step); lastCur = cur; cur += bs*step # _batchRange
    if includeLast and lastCur < stop: yield range(lastCur, stop, step)          # _batchRange
def _batchSliceable(it, bs, includeLast, incl):                                  # _batchSliceable
    cur = 0; n = len(it)                                                         # _batchSliceable
    if incl:                                                                     # _batchSliceable
        while (cur+1)*bs <= n: yield it[cur*bs:(cur+1)*bs+1]; cur += 1           # _batchSliceable
    else:                                                                        # _batchSliceable
        while (cur+1)*bs <= n: yield it[cur*bs:(cur+1)*bs]; cur += 1             # _batchSliceable
    if includeLast:                                                              # _batchSliceable
        ans = it[cur*bs:(cur+1)*bs]                                              # _batchSliceable
        if len(ans) > 0: yield ans                                               # _batchSliceable
class batched(BaseCli):                                                          # batched
    def __init__(self, bs=32, includeLast=False, incl=False):                    # batched
        """Batches the input stream.
Example::

    # returns [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    range(11) | batched(3) | deref()
    # returns [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10]]
    range(11) | batched(3, True) | deref()
    # returns [[0, 1, 2, 3, 4]]
    range(5) | batched(float("inf"), True) | deref()
    # returns []
    range(5) | batched(float("inf"), False) | deref()

    # returns [[0, 1, 2, 3], [3, 4, 5, 6], [6, 7, 8, 9]], includes the first element of the next batch!
    range(11) | batched(3, incl=True) | deref()
    # returns [[0, 1, 2, 3], [3, 4, 5, 6], [6, 7, 8, 9], [9, 10]]
    range(11) | batched(3, True, incl=True) | deref()

Can work well and fast with :class:`torch.Tensor` and :class:`numpy.ndarray`::

    # both returns torch.Tensor of shape (2, 3, 4, 5)
    torch.randn(6, 4, 5) | batched(3)
    torch.randn(7, 4, 5) | batched(3)

Also, if input is a :class:`range`, then to save time, a bunch of other
ranges will be returned, instead of a bunch of lists, for performance::

    # returns [range(0, 3), range(3, 6), range(6, 9)]
    range(11) | batched(3) | toList()

See also: :class:`window`, :class:`batchedTrigger`

:param bs: batch size
:param includeLast: whether to include the last incomplete batch or not
:param incl: whether to have inclusive bounds or not"""                          # batched
        super().__init__(); self.bs = bs; self.includeLast = includeLast; self.incl = incl # batched
    def _all_array_opt(self, it, level):                                         # batched
        if self.includeLast: return NotImplemented                               # batched
        bs = self.bs; a = [slice(None, None, None)]*level; n = it.shape[level] // bs; it = it[(*a, slice(None, n*bs))] # batched
        return it.reshape(*it.shape[:level], n, bs, *it.shape[level+1:])         # batched
    def __ror__(self, it):                                                       # batched
        bs = self.bs; includeLast = self.includeLast; incl = self.incl           # batched
        if bs == float("inf"): return [it] if includeLast else []                # batched
        if not incl and isinstance(it, k1lib.settings.cli.arrayTypes):           # batched
            if (not includeLast) or (it.shape[0]%bs == 0): n = it.shape[0] // bs; it = it[:n*bs]; return it.reshape(n, bs, *it.shape[1:]) # batched
        if not incl and isinstance(it, range): return _batchRange(it, bs, includeLast, incl) # batched
        try: it[0]; len(it); return _batchSliceable(it, bs, includeLast, incl)   # batched
        except: return _batch(it, bs, includeLast, incl)                         # batched
    def _jsF(self, meta):                                                        # batched
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # batched
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.batched({cli.kjs.v(self.bs)}, {cli.kjs.v(self.includeLast)})", fIdx # batched
class batchedTrigger(BaseCli): # batched the data using a trigger column         # batchedTrigger
    def __init__(self, col:int=None, value:"any"=None, delta:float=None, adj:bool=True): # value is to only yield segments that have the specified value # batchedTrigger
        """Like :class:`batched`, will batch rows/elements up but based on
a trigger value instead. See :class:`~k1lib.cli.filt.trigger` for more
discussion. Normal mode::

    data = [[1, "a"], [1, "b"], [2, "c"], [3, "d"], [3, "e"], [1, "f"]]
    # returns [[1, 1], [2], [3, 3], [1]]
    [1, 1, 2, 3, 3, 1] | batchedTrigger()  | deref()
    # throws error, does not make sense to invert search. Can only invert if .value is specified
    [1, 1, 2, 3, 3, 1] | ~batchedTrigger() | deref()
    # returns [[[1, 'a'], [1, 'b']], [[2, 'c']], [[3, 'd'], [3, 'e']], [[1, 'f']]]
    data | batchedTrigger(0) | deref()

In this mode, elements of the specified column that have the same value consecutively
will be grouped together into a batch. Note the difference between this and :class:`groupBy`.
Specific-value mode::

    # returns [[1, 1], [1]]
    [1, 1, 2, 3, 3, 1] | batchedTrigger(value=1)  | deref()
    # returns [[2], [3, 3]], essentially only yield the batches that don't match the specified value
    [1, 1, 2, 3, 3, 1] | ~batchedTrigger(value=1) | deref()
    # returns [[[1, 'a'], [1, 'b']], [[1, 'f']]]
    data | batchedTrigger(0, value=1) | deref()

This mode is pretty much the same as previously, but will only return
the batches whose specified columns match (or not match) a specific value.
Finally, the delta mode::

    # returns [[1, 1.1, 1.2], [2], [3], [3.55, 3.8, 3.9, 4.1], [6], [10]]
    [1, 1.1, 1.2, 2, 3, 3.55, 3.8, 3.9, 4.1, 6, 10] | batchedTrigger(delta=0.5) | deref()
    # returns [[1, 1.1, 1.2], [2], [3], [3.55, 3.8, 3.9], [4.1], [6], [10]]
    [1, 1.1, 1.2, 2, 3, 3.55, 3.8, 3.9, 4.1, 6, 10] | batchedTrigger(delta=0.5, adj=False) | deref()
    # .col also works here, but it's too long to fit here

In contrast to the previous 2 modes, this does not rely on exact values. Instead,
2 consecutive elements are considered to be in the same batch if their difference
is less than the provided ``delta`` value. Within delta mode, there's the parameter
.adj as well. If this is true, then it will measure the delta between the last (3.9)
and current element (4.1), while if .adj=False, then it will measure the delta between
the first element in the group (3.55) and the current element (4.1, which exceeds
delta, which splits into a new batched).

See also: :class:`~k1lib.cli.filt.trigger`

:param col: column of the trigger value. Can be None, int, or list[int]
:param value: if trigger value equals to this, then yields the batch
:param delta: if trigger value changes more than this, then yields the batch
:param adj: "adjacent", only makes sense if ``delta != None``. See example above""" # batchedTrigger
        self.col = col; self.value = value; self.delta = delta; self.adj = adj; self.inverted = False # batchedTrigger
        if value is not None and delta is not None: raise Exception(f"Can't specify precise value and fuzzy values (by specifying delta) at the same time. Does not make sense! Leave either .value or .delta empty") # batchedTrigger
    def __invert__(self):                                                        # batchedTrigger
        if self.value is None: raise Exception("batchedTrigger().value is None! Inverting it would not make sense!") # batchedTrigger
        res = batchedTrigger(self.col, self.value); res.inverted = not self.inverted; return res # batchedTrigger
    def __ror__(self, it):                                                       # batchedTrigger
        col = self.col; value = self.value; delta = self.delta; adj = self.adj; inv = self.inverted # batchedTrigger
        arr = []; lastE = object() # sentinel                                    # batchedTrigger
        # also, the reason for so many if statements because I don't want to have the ifs inside the for loops, cause that's gonna be slow # batchedTrigger
        if value is None and delta is None: # normal mode                        # batchedTrigger
            if col is None:                                                      # batchedTrigger
                for e in it:                                                     # batchedTrigger
                    if e == lastE: arr.append(e)                                 # batchedTrigger
                    else:                                                        # batchedTrigger
                        if len(arr): yield arr                                   # batchedTrigger
                        arr = [e]                                                # batchedTrigger
                    lastE = e                                                    # batchedTrigger
            elif isinstance(col, int):                                           # batchedTrigger
                for row in it:                                                   # batchedTrigger
                    e = row[col]                                                 # batchedTrigger
                    if e == lastE: arr.append(row)                               # batchedTrigger
                    else:                                                        # batchedTrigger
                        if len(arr): yield arr                                   # batchedTrigger
                        arr = [row]                                              # batchedTrigger
                    lastE = e                                                    # batchedTrigger
            else:                                                                # batchedTrigger
                for row in it:                                                   # batchedTrigger
                    e = tuple([row[x] for x in col])                             # batchedTrigger
                    if e == lastE: arr.append(row)                               # batchedTrigger
                    else:                                                        # batchedTrigger
                        if len(arr): yield arr                                   # batchedTrigger
                        arr = [row]                                              # batchedTrigger
                    lastE = e                                                    # batchedTrigger
            if len(arr): yield arr                                               # batchedTrigger
        elif value is not None: # normal mode, but only returns specified value  # batchedTrigger
            if col is None:                                                      # batchedTrigger
                for e in it:                                                     # batchedTrigger
                    if e == lastE: arr.append(e)                                 # batchedTrigger
                    else:                                                        # batchedTrigger
                        if len(arr) and (lastE == value) ^ inv: yield arr        # batchedTrigger
                        arr = [e]                                                # batchedTrigger
                    lastE = e                                                    # batchedTrigger
            elif isinstance(col, int):                                           # batchedTrigger
                for row in it:                                                   # batchedTrigger
                    e = row[col]                                                 # batchedTrigger
                    if e == lastE: arr.append(row)                               # batchedTrigger
                    else:                                                        # batchedTrigger
                        if len(arr) and (lastE == value) ^ inv: yield arr        # batchedTrigger
                        arr = [row]                                              # batchedTrigger
                    lastE = e                                                    # batchedTrigger
            else:                                                                # batchedTrigger
                value = tuple(value)                                             # batchedTrigger
                for row in it:                                                   # batchedTrigger
                    e = tuple([row[x] for x in col])                             # batchedTrigger
                    if e == lastE: arr.append(row)                               # batchedTrigger
                    else:                                                        # batchedTrigger
                        if len(arr) and (lastE == value) ^ inv: yield arr        # batchedTrigger
                        arr = [row]                                              # batchedTrigger
                    lastE = e                                                    # batchedTrigger
            if len(arr) and (lastE == value) ^ inv: yield arr                    # batchedTrigger
        elif delta is not None: # delta mode                                     # batchedTrigger
            lastE = float("-inf"); arr = []                                      # batchedTrigger
            if adj:                                                              # batchedTrigger
                if col is None:                                                  # batchedTrigger
                    for e in it:                                                 # batchedTrigger
                        if abs(e-lastE) < delta: arr.append(e)                   # batchedTrigger
                        else:                                                    # batchedTrigger
                            if len(arr): yield arr                               # batchedTrigger
                            arr = [e]                                            # batchedTrigger
                        lastE = e                                                # batchedTrigger
                elif isinstance(col, int):                                       # batchedTrigger
                    for row in it:                                               # batchedTrigger
                        e = row[col]                                             # batchedTrigger
                        if abs(e-lastE) < delta: arr.append(row)                 # batchedTrigger
                        else:                                                    # batchedTrigger
                            if len(arr): yield arr                               # batchedTrigger
                            arr = [row]                                          # batchedTrigger
                        lastE = e                                                # batchedTrigger
                else: raise Exception("Multiple columns not supported in delta mode, as it doesn't make any sense") # batchedTrigger
            else:                                                                # batchedTrigger
                if col is None:                                                  # batchedTrigger
                    for e in it:                                                 # batchedTrigger
                        if abs(e-lastE) < delta: arr.append(e)                   # batchedTrigger
                        else:                                                    # batchedTrigger
                            if len(arr): yield arr                               # batchedTrigger
                            arr = [e]; lastE = e                                 # batchedTrigger
                elif isinstance(col, int):                                       # batchedTrigger
                    for row in it:                                               # batchedTrigger
                        e = row[col]                                             # batchedTrigger
                        if abs(e-lastE) < delta: arr.append(row)                 # batchedTrigger
                        else:                                                    # batchedTrigger
                            if len(arr): yield arr                               # batchedTrigger
                            arr = [row]; lastE = e                               # batchedTrigger
                else: raise Exception("Multiple columns not supported in delta mode, as it doesn't make any sense") # batchedTrigger
            if len(arr): yield arr                                               # batchedTrigger
        else: raise Exception("Unreachable")                                     # batchedTrigger
nothing = object()                                                               # batchedTrigger
class window(BaseCli):                                                           # window
    def __init__(self, n, newList=False, pad=nothing):                           # window
        """Slides window of size n forward and yields the windows.
Example::

    # returns [[0, 1, 2], [1, 2, 3], [2, 3, 4]]
    range(5) | window(3) | deref()
    # returns [[0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4, None], [4, None, None]]
    range(5) | window(3, pad=None) | deref()

If you are doing strange transformations to the result, like
transposing it, then it might complain that the internal deque
(double-ended queue) mutated during iteration. In that case,
then set ``newList`` to True. It's not True by default because
multiple lists will be created, all of which needs memory
allocation, which will be slower::

    # takes 15ms
    range(100000) | window(100) | ignore()
    # takes 48ms, because of allocating lists
    range(100000) | window(100) | ignore()

See also: :class:`~batched`

:param n: size of the window
:param newList: whether to create a new list out of every window or
    not. If False (default), less robust but faster. If True, more
    robust but slower
:param pad: whether to pad the output stream on the end, so that it
    has the same number of elements as the input stream or not"""                # window
        self.n = n; self.listF = (lambda x: list(x)) if newList else (lambda x: iter(x)) # window
        self.pad = pad; self.padBool = pad is not nothing # why do this? Cause in applyMp, "nothing" takes on multiple identities # window
    def __ror__(self, it):                                                       # window
        n = self.n; pad = self.pad; q = deque([], n); listF = self.listF         # window
        for e in it:                                                             # window
            q.append(e)                                                          # window
            if len(q) == n: yield listF(q); q.popleft()                          # window
        if self.padBool:                                                         # window
            for i in range(n-1): q.append(pad); yield listF(q); q.popleft()      # window
    def _jsF(self, meta):                                                        # window
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # window
        if self.padBool: raise Exception("window._jsF() doesn't support custom pad value yet") # window
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.window({cli.kjs.v(self.n)})", fIdx # window
class groupBy(BaseCli):                                                          # groupBy
    def __init__(self, column:int, separate:bool=False, removeCol:bool=None):    # groupBy
        """Groups table by some column.
Example::

    a = [[2.3, 5],
         [3.4, 2],
         [4.5, 2],
         [5.6, 5],
         [6.7, 1]]
    a | groupBy(1) | deref()

This returns::

    [[[6.7, 1]],
     [[3.4, 2], [4.5, 2]],
     [[2.3, 5], [5.6, 5]]]

Should have O(n log(n)) time complexity. What if ``separate`` is True::

    a | groupBy(1, True) | deref()

This returns::

    [[1, [[6.7]]],
     [2, [[3.4], [4.5]]],
     [5, [[2.3], [5.6]]]]

What if ``removeCol`` is False::

    a | groupBy(1, True, False) | deref()

This returns::

    [[1, [[6.7, 1]]],
     [2, [[3.4, 2], [4.5, 2]]],
     [5, [[2.3, 5], [5.6, 5]]]]

There's another perspective and way to think about this operation. A lot of
libraries (like pandas) expect the uncompressed, "flat" version (the variable
``a`` in the examples above). But throughout my time using cli, the grouped
version (separate=True) is usually much more useful and amenable to
transformations. It also occupies less memory too, as the columns with duplicated
elements are deleted.

So, you can sort of think :class:`groupBy` is converting pandas dataframes
into a more easily digestible form. But because the prevalence of those libraries,
after doing all the transformations you want, sometimes it's necessary to flatten
it again, which :class:`ungroup` does.

If you want to group text lines by some pattern im them, :class:`~k1lib.cli.grep.grep`
with ``sep=True`` might be better for you.

:param column: which column to group by
:param separate: whether to separate out the column to sort of form a dict or not. See example
:param removeCol: whether to remove the grouped-by column. Defaults to True if
    ``separate=True``, and False if ``separate=False``"""                        # groupBy
        self.column = column; self.separate = separate                           # groupBy
        if removeCol is None: removeCol = separate # if separate, then remove cols, else don't do it # groupBy
        self.removeCol = removeCol                                               # groupBy
    def __ror__(self, it):                                                       # groupBy
        it = [list(e) for e in it]; c = self.column; separate = self.separate; removeCol = self.removeCol # groupBy
        it = it | cli.sort(c, False); sentinel = object(); it = iter(it); a = [next(it, sentinel)] # groupBy
        if a[0] is sentinel: return                                              # groupBy
        v = a[0][c]                                                              # groupBy
        try:                                                                     # groupBy
            while True:                                                          # groupBy
                e = next(it)                                                     # groupBy
                if e[c] == v: a.append(e)                                        # groupBy
                else:                                                            # groupBy
                    if removeCol: a = a | ~cli.cut(c)                            # groupBy
                    yield [v, a] if separate else a; a = [e]; v = a[0][c]        # groupBy
        except StopIteration:                                                    # groupBy
            if len(a) > 0:                                                       # groupBy
                if removeCol: a = a | ~cli.cut(c)                                # groupBy
                yield [v, a] if separate else a                                  # groupBy
    def _jsF(self, meta):                                                        # groupBy
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # groupBy
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.groupBy({cli.kjs.v(self.column)}, {cli.kjs.v(self.separate)}, {cli.kjs.v(self.removeCol)})", fIdx # groupBy
class ungroup(BaseCli):                                                          # ungroup
    def __init__(self, single=True, begin=True, insertCol:bool=True):            # ungroup
        """Ungroups things that were grouped using a specific mode of
:class:`groupBy`. Particularly useful to transform some complex data
structure into a flat dataframe so that you can plug into pandas. Example::

    # returns [[3, 1.2], [3, 3.4], [5, 6], [5, 8], [5, 11]]
    [[3, [1.2, 3.4]], [5, [6, 8, 11]]] | ungroup() | deref()
    # returns [[3, 1.2], [3, 3.4], [5, 6], [5, 8], [5, 11]]
    [[3, [[1.2], [3.4]]], [5, [[6], [8], [11]]]] | ungroup(False) | deref()
    # returns [[1.2, 3], [3.4, 3], [6, 5], [8, 5], [11, 5]]
    [[3, [1.2, 3.4]], [5, [6, 8, 11]]] | ungroup(begin=False) | deref()

:param single: whether the table in each group has a single column or not
:param begin: whether to insert the column at the beginning or at the end.
    Only works if ``insertCol`` is True
:param insertCol: whether to insert the column into the table or not"""          # ungroup
        self.single = single; self.begin = begin; self.insertCol = insertCol     # ungroup
    def __ror__(self, it):                                                       # ungroup
        preprocess = cli.apply(cli.wrapList().all(), 1) if self.single else cli.iden(); begin = self.begin # ungroup
        if self.insertCol: return it | preprocess | ~cli.apply(lambda x, arr: arr | cli.insert(x, begin).all()) | cli.joinStreams() # ungroup
        else: return it | preprocess | cli.cut(1) | cli.joinStreams()            # ungroup
    def _jsF(self, meta):                                                        # ungroup
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # ungroup
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.ungroup({cli.kjs.v(self.single)}, {cli.kjs.v(self.begin)}, {cli.kjs.v(self.insertCol)})", fIdx # ungroup
class insertColumn(BaseCli):                                                     # insertColumn
    def __init__(self, column:List[Any], begin=True, fill=""):                   # insertColumn
        """Inserts a column at beginning or end.
Example::

    # returns [['a', 1, 2], ['b', 3, 4]]
    [[1, 2], [3, 4]] | insertColumn(["a", "b"]) | deref()
    # returns [[1, 2, 'a'], [3, 4, 'b']]
    [[1, 2], [3, 4]] | insertColumn(["a", "b"], begin=False) | deref()"""        # insertColumn
        self.column = column; self.begin = begin; self.fill = fill               # insertColumn
        self._f = transpose.wrap(insert(column, begin=begin), fill=fill)         # insertColumn
    def __ror__(self, it): return it | self._f                                   # insertColumn
    def _jsF(self, meta): return self._f._jsF(meta)                              # insertColumn
class insertIdColumn(BaseCli):                                                   # insertIdColumn
    def __init__(self, table=False, begin=True):                                 # insertIdColumn
        """Inserts an id column at the beginning (or end).
Example::

    # returns [[0, 'a', 2], [1, 'b', 4]]
    [["a", 2], ["b", 4]] | insertIdColumn(True) | deref()
    # returns [[0, 'a'], [1, 'b']]
    "ab" | insertIdColumn()

This has a shorter alias called :class:`insId`

:param table: if False, then insert column to an Iterator[str], else treat
    input as a full fledged table"""                                             # insertIdColumn
        self.table = table; self.begin = begin                                   # insertIdColumn
    def __ror__(self, it):                                                       # insertIdColumn
        if self.table:                                                           # insertIdColumn
            if self.begin: return ([i, *e] for i, e in enumerate(it))            # insertIdColumn
            else: return ([*e, i] for i, e in enumerate(it))                     # insertIdColumn
        else:                                                                    # insertIdColumn
            if self.begin: return ([i, e] for i, e in enumerate(it))             # insertIdColumn
            else: return ([e, i] for i, e in enumerate(it))                      # insertIdColumn
    def _jsF(self, meta):                                                        # insertIdColumn
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # insertIdColumn
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.insertIdColumn({cli.kjs.v(self.table)}, {cli.kjs.v(self.begin)})", fIdx # insertIdColumn
insId = insertIdColumn                                                           # insertIdColumn
def unsqueeze(dim:int=0):                                                        # unsqueeze
    """Unsqueeze input iterator.
Example::

    t = [[1, 2], [3, 4], [5, 6]]
    # returns (3, 2)
    t | shape()
    # returns (1, 3, 2)
    t | unsqueeze(0) | shape()
    # returns (3, 1, 2)
    t | unsqueeze(1) | shape()
    # returns (3, 2, 1)
    t | unsqueeze(2) | shape()

Behind the scenes, it's really just ``wrapList().all(dim)``, but the "unsqueeze" name
is a lot more familiar. Also note that the inverse operation "squeeze" is sort of
``item().all(dim)``, if you're sure that this is desirable::

    t = [[1, 2], [3, 4], [5, 6]]
    # returns (3, 2)
    t | unsqueeze(1) | item().all(1) | shape()"""                                # unsqueeze
    return cli.wrapList().all(dim)                                               # unsqueeze
def oUnsqueeze(cs, ts, metadata): # reminder: if change this then also change the example in llvm.rst # oUnsqueeze
    a = cs[0]; t = ts[0]; i = 0;                                                 # oUnsqueeze
    if not isinstance(t, tArrayTypes): return None                               # oUnsqueeze
    while isinstance(a, cli.apply) and a.normal: i += 1; a = a.f                 # oUnsqueeze
    if not isinstance(a, cli.wrapList): return None                              # oUnsqueeze
    t = t.__class__(t.child, t.rank+1 if t.rank is not None else None)           # oUnsqueeze
    if isinstance(t, tNpArray): return [cli.aS(lambda x: np.expand_dims(x, i)).hint(t)] # oUnsqueeze
    else: return [cli.aS(lambda x: x.unsqueeze(i)).hint(t)]                      # oUnsqueeze
tOpt.addPass(oUnsqueeze, [cli.apply], 4)                                         # oUnsqueeze
class count(BaseCli):                                                            # count
    def __init__(self, max=False):                                               # count
        """Finds unique elements and returns a table with [frequency, value, percent]
columns. Example::

    # returns [[1, 'a', '33%'], [2, 'b', '67%']]
    ['a', 'b', 'b'] | count() | deref()
    # returns [[1, 'a', '50%'], [2, 'b', '100%']]
    ['a', 'b', 'b'] | count(max=True) | deref()

:param max: if True, percentages are of the max value, else they're of the sum
"""                                                                              # count
        super().__init__(); self._max = max                                      # count
    def _typehint(self, inp):                                                    # count
        i = tAny()                                                               # count
        if isinstance(inp, tListIterSet): i = inp.child                          # count
        return tIter(tCollection(int, i, str))                                   # count
    def __ror__(self, it:Iterator[str]):                                         # count
        it = it | cli.apply(lambda row: (tuple(row) if isinstance(row, list) else row)) # count
        c = Counter(it); s = max(c.values()) if self._max else sum(c.values())   # count
        return [[v, k, f"{round(100*v/s)}%"] for k, v in c.items()] # has to scan through the entire thing anyway, which is long enough already, so just turn it into a list # count
    @staticmethod                                                                # count
    def join():                                                                  # count
        """Joins multiple counts together.
Example::

    # returns [[2, 'a', '33%'], [4, 'b', '67%']]
    ['a', 'b', 'b'] | repeat(2) | applyMp(count() | deref()) | count.join() | deref()

This is useful when you want to get the count of a really long list/iterator using multiple cores""" # count
        def inner(counts):                                                       # count
            values = defaultdict(lambda: 0)                                      # count
            for _count in counts:                                                # count
                if _count is None: continue                                      # count
                for v, k, *_ in _count:                                          # count
                    values[k] += v                                               # count
            s = values.values() | cli.toSum()                                    # count
            return [[v, k, f"{round(100*v/s)}%"] for k, v in values.items()]     # count
        return cli.applyS(inner)                                                 # count
    def _jsF(self, meta):                                                        # count
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # count
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.count()", fIdx          # count
class hist(BaseCli):                                                             # hist
    def __init__(self, bins:int=30, dropZero:bool=False):                        # hist
        """Bins a long 1d array. Effectively creating a historgram, without
actually plotting it. Example::

    np.random.randn(1000) | hist(5)

That returns something like::

    (array([-2.31449761, -1.17406889, -0.03364017,  1.10678854,  2.24721726]),
     array([ 41, 207, 432, 265,  55]),
     1.1404287156493986)

This format goes with :meth:`~matplotlib.pyplot.bar` directly like this::

    np.random.randn(1000) | hist(10) | ~aS(plt.bar)

If you have tons of data that's handled in multiple processes, but you want to
get an overall histogram, you can do something like this::

    # bad solution, runs slow, accurate
    fileNames | applyMp(cat() | toFloat() | aS(list)) | joinStreams() | hist() | ~aS(plt.bar)
    # good solution, runs fast, slightly inaccurate
    fileNames | applyMp(cat() | toFloat() | hist(300)) | hist.join() | ~aS(plt.bar)

Let's say in each process, you have 10M records, and that you have 1000 processes in total.
In the first solution, you transfer all records (10B records total) to a single process,
then calculate the histogram of them. The transfer overhead is going to be absolutely
enourmous, as well as the computation. This really defeats the purpose of doing
multiprocessing.

In the second solution, you "convert" 10M records into 600 numbers for each process,
which scales up to 600k numbers for all processes. Although big, but certainly
manageable with current hardware. So the data transfer cost is not a lot at all. The
histogram merging part also executes relatively fast, as it only creates an internal
array of 3M length. See over :meth:`hist.join` for param details

:param bins: how many bins should the histogram has?
:param dropZero: if True, will eliminate buckets that has no element in them"""  # hist
        self.bins = bins; self.dropZero = dropZero                               # hist
    def __ror__(self, it):                                                       # hist
        if not isinstance(it, settings.arrayTypes): it = list(it)                # hist
        y, x = np.histogram(it, bins=self.bins)                                  # hist
        delta = x[1] - x[0]; x = (x[1:] + x[:-1])/2                              # hist
        if self.dropZero: idx = y > 0; return x[idx], y[idx], delta              # hist
        return x, y, delta                                                       # hist
    @staticmethod                                                                # hist
    def join(scale:float=1e4, bins:int=None, log:bool=True, xlog:bool=False):    # hist
        """Joins multiple histograms together.
Example::

    a = np.random.randn(1000); b = np.random.randn(1000)+3
    [a, b] | joinStreams() | hist() | head(2) | ~aS(plt.plot) # ---------------------------------- Ground truth
    [a, b] | hist(300).all() | hist.join(scale=1e4) | head(2) | ~aS(plt.plot) # ------------------ Log joining
    [a, b] | hist(300).all() | hist.join(scale=1e4, log=False) | head(2) | ~aS(plt.plot, ".") # -- Linear joining
    plt.legend(["Ground truth", "Log", "Linear"]); plt.grid(True); plt.ylabel("Frequency"); plt.xlabel("Value");

This results in this:

.. image:: ../images/hist1.png

As you can see, this process is only approximate, but is accurate enough in everyday
use. If you are a normal user, then this is probably enough. However, if you're a
mathematician and really care about the accuracy of this, read on.

.. admonition:: Performance vs accuracy

    As mentioned in :class:`hist`, joining histograms from across processes can really speed
    things up big time. But the joining process is complicated, with multiple parameters
    and different tradeoffs for each config. In this example, scale = 1e4, bins = 30,
    OG bins = 300, log = True. "OG bins" is the number of bins coming into :meth:`hist.join`

    To get the best accuracy possible, you should set scale and OG bins high. If better
    performance is desired, you should first lower scale, then lower OG bins, then finally lower
    bins.

.. admonition:: Log scale

    Take a look at this piece of code::

        a, b = np.random.randn(1000)*1, np.random.randn(3000000)*0.3+3
        [a, b] | joinStreams() | hist() | head(2) | ~aS(plt.plot)
        [a, b] | hist(300).all() | hist.join(scale=1e4) | head(2) | ~aS(plt.plot)
        [a, b] | hist(300).all() | hist.join(scale=1e4, log=False) | head(2) | ~aS(plt.plot, ".")
        plt.yscale("log"); plt.legend(["Ground truth", "Log", "Linear"]); plt.grid(True); plt.ylabel("Frequency"); plt.xlabel("Value");

    This results in:

    .. image:: ../images/hist2.png

    This shows how log mode is generally better than linear mode when the frequencies span
    across multiple orders of magnitude. So why not delete linear mode directly? Well, I have
    not formally proved that log scale case fully covers the linear case, although in practice
    it seems so. So just to be cautious, let's leave it in

.. admonition:: Scale

    The setup is just like in the "Log scale" section, but with scale = 1e3 instead of the default 1e4:

    .. image:: ../images/hist3.png

    Remember that the higher the scale, the more accurate, but also the longer it runs. If
    the difference in high and low frequencies are bigger than scale, then the low
    frequency values are dropped.

:param scale: how big a range of frequencies do we want to capture?
:param bins: output bins. If not specified automatically defaults to 1/10 the
    original number of bins
:param log: whether to transform everything to log scale internally on the y axis
:param xlog: whether to transform everything to log scale internally on the x axis""" # hist
        def inner(it):                                                           # hist
            it = it | (cli.apply(cli.apply(math.log), 0) if xlog else cli.iden()) | cli.deref() # hist
            _bins = bins if bins is not None else it | cli.cut(0) | cli.shape(0).all() | cli.toMean() | cli.op()/10 | cli.aS(int) # hist
            maxY = max(it | cli.cut(1) | cli.toMax().all() | cli.toMax() | cli.op()/scale, 1e-9) # hist
            if log: it = it | cli.apply(lambda x: np.log(x+1e-9)-math.log(maxY) | cli.aS(np.exp) | cli.aS(np.round) | cli.op().astype(int), 1) | cli.deref() # hist
            else: it = it | cli.apply(lambda y: (y/maxY).astype(int), 1) | cli.deref() # hist
            x, y, delta = it | cli.cut(0, 1) | cli.transpose().all() | cli.joinStreams() | ~cli.apply(lambda a,b: [a]*b) | cli.joinStreams() | cli.aS(list) | cli.aS(np.array) | hist(_bins) # hist
            return x | ((cli.apply(math.exp) | cli.deref()) if xlog else cli.iden()), y*maxY, delta # hist
        return cli.aS(inner)                                                     # hist
def _permuteGen(row, pers):                                                      # _permuteGen
    row = list(row); return (row[i] for i in pers)                               # _permuteGen
class permute(BaseCli):                                                          # permute
    def __init__(self, *permutations:List[int]):                                 # permute
        """Permutes the columns. Acts kinda like :meth:`torch.Tensor.permute`.
Example::

    # returns [['b', 'a'], ['d', 'c']]
    ["ab", "cd"] | permute(1, 0) | deref()"""                                    # permute
        super().__init__(); self.permutations = permutations                     # permute
    def __ror__(self, it:Iterator[str]):                                         # permute
        p = self.permutations                                                    # permute
        for row in it: yield _permuteGen(row, p)                                 # permute
    def _jsF(self, meta):                                                        # permute
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # permute
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.permute({cli.kjs.vs(self.permutations) | cli.join(', ')})", fIdx # permute
class AA_(BaseCli):                                                              # AA_
    def __init__(self, *idxs:List[int], wraps=False):                            # AA_
        """Returns 2 streams, one that has the selected element, and the other
the rest. Example::

    # returns [5, [1, 6, 3, 7]]
    [1, 5, 6, 3, 7] | AA_(1)
    # returns [[5, [1, 6, 3, 7]]]
    [1, 5, 6, 3, 7] | AA_(1, wraps=True)

You can also put multiple indexes through::

    # returns [[1, [5, 6]], [6, [1, 5]]]
    [1, 5, 6] | AA_(0, 2)

If you don't specify anything, then all indexes will be sliced::

    # returns [[1, [5, 6]], [5, [1, 6]], [6, [1, 5]]]
    [1, 5, 6] | AA_()

As for why the strange name, think of this operation as "AĀ". In statistics,
say you have a set "A", then "not A" is commonly written as A with an overline
"Ā". So "AA\_" represents "AĀ", and that it first returns the selection A.

:param wraps: if True, then the first example will return [[5, [1, 6, 3, 7]]]
    instead, so that A has the same signature as Ā"""                            # AA_
        super().__init__(); self.idxs = idxs; self.wraps = wraps                 # AA_
    def __ror__(self, it:List[Any]) -> List[List[List[Any]]]:                    # AA_
        super().__ror__(it); idxs = self.idxs; it = list(it)                     # AA_
        if len(idxs) == 0: idxs = range(len(it))                                 # AA_
        def gen(idx):                                                            # AA_
            return [it[idx], [v for i, v in enumerate(it) if i != idx]]          # AA_
        if not self.wraps and len(idxs) == 1: return gen(idxs[0])                # AA_
        return [gen(idx) for idx in idxs]                                        # AA_
class peek(BaseCli):                                                             # peek
    def __init__(self):                                                          # peek
        """Returns (firstRow, iterator). This sort of peaks at the first row,
to potentially gain some insights about the internal formats. The returned
iterator is not tampered. Example::

    e, it = iter([[1, 2, 3], [1, 2]]) | peek()
    print(e) # prints "[1, 2, 3]"
    s = 0
    for e in it: s += len(e)
    print(s) # prints "5", or length of 2 lists

You kinda have to be careful about handling the ``firstRow``, because you might
inadvertently alter the iterator::

    e, it = iter([iter(range(3)), range(4), range(2)]) | peek()
    e = list(e) # e is [0, 1, 2]
    list(next(it)) # supposed to be the same as `e`, but is [] instead

The example happens because you have already consumed all elements of the first
row, and thus there aren't any left when you try to call ``next(it)``."""        # peek
        super().__init__()                                                       # peek
    def __ror__(self, it:Iterator[Any]) -> Tuple[Any, Iterator[Any]]:            # peek
        it = iter(it); sentinel = object(); row = next(it, sentinel)             # peek
        if row is sentinel: return None, []                                      # peek
        def gen(): yield row; yield from it                                      # peek
        return row, gen()                                                        # peek
class peekF(BaseCli):                                                            # peekF
    def __init__(self, f:Union[BaseCli, Callable[[Any], Any]]):                  # peekF
        r"""Similar to :class:`peek`, but will execute ``f(row)`` and
return the input Iterator, which is not tampered. Example::

    it = lambda: iter([[1, 2, 3], [1, 2]])
    # prints "[1, 2, 3]" and returns [[1, 2, 3], [1, 2]]
    it() | peekF(lambda x: print(x)) | deref()
    # prints "1\n2\n3"
    it() | peekF(headOut()) | deref()"""                                         # peekF
        super().__init__(fs=[f]); self.f = f                                     # peekF
    def __ror__(self, it:Iterator[Any]) -> Iterator[Any]:                        # peekF
        it = iter(it); sentinel = object(); row = next(it, sentinel)             # peekF
        if row is sentinel: return []                                            # peekF
        def gen(): yield row; yield from it                                      # peekF
        self.f(row); return gen()                                                # peekF
settings.add("repeat", k1lib.Settings().add("infBs", 100, "if dealing with infinite lists, how many elements at a time should be processed?"), "settings related to repeat() and repeatFrom()") # peekF
class repeat(BaseCli):                                                           # repeat
    def __init__(self, limit:int=None):                                          # repeat
        """Yields a specified amount of the passed in object. If you intend
to pass in an iterator, then make a list out of it first, as second copy of
iterator probably won't work as you will have used it the first time. Example::

    # returns [[1, 2, 3], [1, 2, 3], [1, 2, 3]]
    [1, 2, 3] | repeat(3) | deref()

If ``limit`` is not specified, then this will repeat indefinitely, which may or may
not cause a problem downstream. It could be that some downstream operators want to
get the full view of all elements in the input iterator, but because it's an infinite
iterator, it will never complete executing and get the full view that it wants::

    [1, 2] |  repeat() | deref()  | headOut() # will hang indefinitely, fill RAM up and crash after a while
    [1, 2] | (repeat() | deref()) | headOut() # will not hang, and actually print out the first 10 lines, each line is "[1, 2]"

The solution is to just make a group that starts with ``repeat()``. Then ``repeat()``
can capture all downstream operations, and actually batch up the infinite iterator,
pass each batch into the downstream operations, and merge the results together. The
batch size can be controlled by ``settings.cli.repeat.infBs = 200`` if you need it
for some reason.

See also: :meth:`repeatF`

:param repeat: if None, then repeats indefinitely"""                             # repeat
        super().__init__(capture=True); self.limit = limit                       # repeat
    def _typehint(self, inp): return tIter(inp)                                  # repeat
    def _all_array_opt(self, it, level):                                         # repeat
        if self.limit is None or self.limit == float("inf"): return NotImplemented # repeat
        sh = it | cli.shape(); s = slice(None, None, None)                       # repeat
        it = it[(*[s]*level, None)]; newSh = [*sh[:level], self.limit, *sh[level:]] # repeat
        if isinstance(it, np.ndarray): return np.broadcast_to(it, newSh) | self.capturedSerial.all(level) # repeat
        elif hasTorch and isinstance(it, torch.Tensor): return it.expand(newSh) | self.capturedSerial.all(level) # repeat
        return NotImplemented                                                    # repeat
    def __ror__(self, o:Any) -> Iterator[Any]:                                   # repeat
        limit = self.limit if self.limit != None else k1lib.settings.cli.inf; ser = self.capturedSerial # repeat
        def gen():                                                               # repeat
            for i in itertools.count():                                          # repeat
                if i >= limit: break                                             # repeat
                yield o                                                          # repeat
        if limit < float("inf"): return self._all_array_opt(o, 0) if isinstance(o, settings.arrayTypes) else (gen() | ser) # repeat
        return (gen() | ser) if len(self.capturedClis) == 0 else gen() | cli.batched(settings.repeat.infBs) | cli.apply(ser) | cli.joinStreams() # repeat
    def _jsF(self, meta):                                                        # repeat
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # repeat
        if self.limit is None: raise Exception("repeat._jsF() can't repeat the input array infinitely many times, as there's no iterator concept in JS") # repeat
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.repeat({cli.kjs.v(self.limit)})", fIdx # repeat
class _repeatF(BaseCli):                                                         # _repeatF
    def __init__(self, limit, kwargs): super().__init__(capture=True); self.limit = limit; self.kwargs = kwargs # _repeatF
    def __ror__(self, f):                                                        # _repeatF
        f = fastF(f); limit = self.limit; kwargs = self.kwargs; ser = self.capturedSerial # _repeatF
        limit = limit if limit != None else k1lib.settings.cli.inf               # _repeatF
        def gen():                                                               # _repeatF
            if len(kwargs) == 0:                                                 # _repeatF
                for i in itertools.count():                                      # _repeatF
                    if i >= limit: break                                         # _repeatF
                    yield f()                                                    # _repeatF
            else:                                                                # _repeatF
                for i in itertools.count():                                      # _repeatF
                    if i >= limit: break                                         # _repeatF
                    yield f(**kwargs)                                            # _repeatF
        if limit < float("inf"): return gen() | ser                              # _repeatF
        return (gen() | ser) if len(self.capturedClis) == 0 else gen() | cli.batched(settings.repeat.infBs) | cli.apply(ser) | cli.joinStreams() # _repeatF
def repeatF(f=None, limit:int=None, **kwargs):                                   # repeatF
    """Yields a specified amount generated by a specified function.
Example::

    repeatF(lambda: 4, 3) | deref()          # returns [4, 4, 4]
    (lambda: 4) | repeatF(limit=3) | deref() # returns [4, 4, 4]
    repeatF(lambda: 4) | head() | shape(0)   # returns 10

    f = lambda a: a+2
    repeatF(f, 3, a=6) | deref()        # returns [8, 8, 8]
    f | repeatF(limit=3, a=6) | deref() # returns [8, 8, 8]

See :class:`repeat` for a discussion on how to deal with infinite iterators.
If you want to do something like that, then you have to pipe in f instead of
specifying f in the params directly.

:param f: function to repeat. If None, then you can pipe the function in
:param limit: if None, then repeats indefinitely
:param kwargs: extra keyword arguments that you can pass into the function

See also: :class:`repeatFrom`"""                                                 # repeatF
    return _repeatF(limit, kwargs) if f is None else f | _repeatF(limit, kwargs) # repeatF
class repeatFrom(BaseCli):                                                       # repeatFrom
    def __init__(self, limit:int=None):                                          # repeatFrom
        """Yields from a list. If runs out of elements, then do it again for
``limit`` times. Example::

    # returns [1, 2, 3, 1, 2]
    [1, 2, 3] | repeatFrom() | head(5) | deref()
    # returns [1, 2, 3, 1, 2, 3]
    [1, 2, 3] | repeatFrom(2) | deref()

See :class:`repeat` for a discussion on how to deal with infinite iterators

.. note::

    For advanced users who wants to modify the resulting stream mid-way, read this section

    Because this reuses elements inside the input iterator, it's necessary
    that the input feels like a list and not an iterator. So in order to make
    this work::

        # returns [1, 2, 3, 1, 2, 3]
        iter([1, 2, 3]) | repeatFrom(2) | deref()

    It's necessary to turn the input iterator into a list. However, sometimes you
    may want to update the input iterator values, so as to make things extra
    dynamic, like this::

        l = [1, 2, 3]
        def g(): yield from l; yield from l
        def h():
            for i, e in enumerate(g()):
                if i == 3: l.append(5) # modifies the list mid-way
                yield e

        h() | deref() # returns [1, 2, 3, 1, 2, 3, 5]

    But if you do this, it wouldn't work::

        l = [1, 2, 3]
        def h():
            for i, e in enumerate(iter(l) | repeatFrom(2)):
                if i == 3: l.append(5)
                yield e
        h() | deref() # returns [1, 2, 3, 1, 2, 3]

    This is because internally, :class:`repeatFrom` turns the iterator into a
    list, and continues yielding from that list, and thus won't use the updated
    values. To do it, you have to make the input feels like a list (can get length)::

        l = [1, 2, 3]
        def h():
            for i, e in enumerate(l | repeatFrom(2)): # ---------------- changed section
                if i == 3: l.append(5)
                yield e
        h() | deref() # returns [1, 2, 3, 1, 2, 3, 5]

:param limit: if None, then repeats indefinitely"""                              # repeatFrom
        super().__init__(capture=True); self.limit = limit                       # repeatFrom
    def _typehint(self, inp):                                                    # repeatFrom
        i = tAny()                                                               # repeatFrom
        if isinstance(inp, tListIterSet): i = inp.child                          # repeatFrom
        if isinstance(inp, tArrayTypes): i = inp                                 # repeatFrom
        return tIter(i)                                                          # repeatFrom
    def _all_array_opt(self, it, level:int):                                     # repeatFrom
        if self.limit is None or self.limit == float("inf"): return NotImplemented # repeatFrom
        return it | (repeat(self.limit) | joinStreams()).all(level) | self.capturedSerial.all(level) # repeatFrom
    def __ror__(self, it:Iterator[Any]) -> Iterator[Any]:                        # repeatFrom
        limit = self.limit or k1lib.settings.cli.inf; ser = self.capturedSerial  # repeatFrom
        def gen(it):                                                             # repeatFrom
            try: len(it)                                                         # repeatFrom
            except: it = list(it)                                                # repeatFrom
            for i in itertools.count():                                          # repeatFrom
                if i >= limit: break                                             # repeatFrom
                yield from it                                                    # repeatFrom
        if limit < float("inf"): return self._all_array_opt(it, 0) if isinstance(it, settings.arrayTypes) else (gen(it) | ser) # repeatFrom
        return (gen(it) | ser) if len(self.capturedClis) == 0 else gen(it) | cli.batched(settings.repeat.infBs) | cli.apply(ser) | cli.joinStreams() # repeatFrom
    def _jsF(self, meta):                                                        # repeatFrom
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # repeatFrom
        if self.limit is None: raise Exception(f"repeatFrom._jsF() can't repeat the input array infinitely many times, as there's no iterator concept in JS") # repeatFrom
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.repeatFrom({cli.kjs.v(self.limit)})", fIdx # repeatFrom
def oneHotRow(i, n): ans = [0]*n; ans[i] = 1; return ans                         # oneHotRow
class oneHot(BaseCli):                                                           # oneHot
    _groups = {}                                                                 # oneHot
    def __init__(self, col, n:int=0, group:str=None, sep:bool=False):            # oneHot
        """One-hot encode some column in a table.
Example::

    a = [
     [1, 2, "A"],
     [3, 4, "B"],
     [5, 6, "C"]]
    b = [
     [7, 8, "A"],
     [9, 10, "B"],
     [11, 12, "B"]]
    [*a, *b] | oneHot(2) | deref()
    [*a, *b] | oneHot(2, 3, "abcd") | deref()

Last 2 statements both return this::

    [[1, 2, 1, 0, 0],
     [3, 4, 0, 1, 0],
     [5, 6, 0, 0, 1],
     [7, 8, 1, 0, 0],
     [9, 10, 0, 1, 0],
     [11, 12, 0, 1, 0]]

You can also separate the encoded column out like this::

    [*a, *b] | oneHot(2, sep=True) | deref()

Which returns this::

    [[1, 2, [1, 0, 0]],
     [3, 4, [0, 1, 0]],
     [5, 6, [0, 0, 1]],
     [7, 8, [1, 0, 0]],
     [9, 10, [0, 1, 0]],
     [11, 12, [0, 1, 0]]]

The natural way to do this is to use with without ``n`` and ``group`` parameters.
But sometimes, your one hot encoding is spreaded across multiple datasets in
multiple dataloaders, and so the order and length of the encoding might not be
the same, which will mess up your training process.

That's why, you can specify ``group``, which will share encoding information
across all :class:`oneHot` clis that have the same group name. If you choose to
do this then you have to also specify what's the size of the encoding, because
the cli can't really infer the size when it potentially has not seen all the data
right?

:param col: which column one hot encode and expand into
:param n: (optional) total number of different elements
:param group: (optional) group name
:param sep: (optional) whether to separate the variable out into its own list""" # oneHot
        self.col = col; self.n = n; self.group = group; self.sep = sep           # oneHot
        if (n != 0 and group is not None) and (n == 0 or group is None):         # oneHot
            raise Exception("You have to specify both `n` and `group` at the same time if you want to use them") # oneHot
        if group is not None:                                                    # oneHot
            if group not in oneHot._groups:                                      # oneHot
                oneHot._groups[group] = dict()                                   # oneHot
            self.d = oneHot._groups[group]                                       # oneHot
        else: self.d = dict()                                                    # oneHot
    def _typehint(self, inp):                                                    # oneHot
         # TODO                                                                  # oneHot
        if isinstance(inp, tListIterSet):                                        # oneHot
            if isinstance(inp.child, tListIterSet):                              # oneHot
                pass                                                             # oneHot
            elif isinstance(inp.child, tListIterSet):                            # oneHot
                pass                                                             # oneHot
            pass                                                                 # oneHot
        return tIter(tAny())                                                     # oneHot
    def __ror__(self, it):                                                       # oneHot
        c = self.col; d = self.d; n = self.n; sep = self.sep                     # oneHot
        if n == 0:                                                               # oneHot
            it = it | cli.deref(2); n = it | cli.cut(c) | cli.aS(set) | cli.shape(0) # oneHot
        for row in it:                                                           # oneHot
            e = row[c]                                                           # oneHot
            try: e[0]; len(e)                                                    # oneHot
            except: e = list(e)                                                  # oneHot
            if e not in d: d[e] = oneHotRow(len(d), n)                           # oneHot
            if sep: yield [*row[:c], d[e], *row[c+1:]]                           # oneHot
            else:   yield [*row[:c], *d[e], *row[c+1:]]                          # oneHot
        return                                                                   # oneHot
        _d = it | cli.cut(c) | cli.aS(set) | cli.sort(None, False) | cli.deref(); n = len(_d) # oneHot
        _d = _d | insertIdColumn(begin=False) | cli.apply(cli.aS(oneHotRow, n), 1) | transpose() | cli.toDict() # oneHot
        for row in it: yield [*row[:c], *_d[row[c]], *row[c+1:]]                 # oneHot
class latch(BaseCli):                                                            # latch
    def __init__(self, timeline, before=True, startTok=None, endTok=None):       # latch
        """Latches 1 timeline (B) onto another timeline (A).
Example::

    a = [[1, "a"], [2, "b"], [3, "c"], [4, "d"], [5, "e"], [6, "f"], [7, "g"], [8, "h"], [9, "i"]]
    b = [[2.1, "2.1"], [3.1, "3.1"]]
    c = [[4.5, "4.5"], [7.3, "7.3"]]

    a[:5] | latch(b)        | deref() # returns [[1, 'a', None], [2, 'b', None], [3, 'c', '2.1'], [4, 'd', '3.1'], [5, 'e', '3.1']]
    a[:5] | latch(b, False) | deref() # returns [[1, 'a', '2.1'], [2, 'b', '2.1'], [3, 'c', '3.1'], [4, 'd', None], [5, 'e', None]]
    b     | latch(a)        | deref() # returns [[2.1, '2.1', 'b'], [3.1, '3.1', 'c']]
    b     | latch(a, False) | deref() # returns [[2.1, '2.1', 'c'], [3.1, '3.1', 'd']]

    # latching timelines that don't overlap
    b | latch(c)        | deref() # returns [[2.1, '2.1', None], [3.1, '3.1', None]]
    b | latch(c, False) | deref() # returns [[2.1, '2.1', '4.5'], [3.1, '3.1', '4.5']]
    c | latch(b)        | deref() # returns [[4.5, '4.5', '3.1'], [7.3, '7.3', '3.1']]
    c | latch(b, False) | deref() # returns [[4.5, '4.5', None], [7.3, '7.3', None]]

    # multiple latches also work fine
    a | latch(b) | latch(c) | deref()

Final command returns this::

    [[1, 'a', None, None],
     [2, 'b', None, None],
     [3, 'c', '2.1', None],
     [4, 'd', '3.1', None],
     [5, 'e', '3.1', '4.5'],
     [6, 'f', '3.1', '4.5'],
     [7, 'g', '3.1', '4.5'],
     [8, 'h', '3.1', '7.3'],
     [9, 'i', '3.1', '7.3']]

So essentially, there're 2 timelines, "a" and "b". "a" is usually more densely
packed than "b". The goal is to merge these 2 timelines together, trying to insert
b's data points at various points inside a. Try to figure out the pattern and
behavior from the example.

This cli makes several assumptions:

- All timelines inside the argument (if ``a | latch(b)``, then I'm talking about b) have only 2 columns: (time, data)
- Timelines piped into (if ``a | latch(b)``, then I'm talking about a) can have multiple columns. First column has to be time
- All timelines are sorted by time already, to avoid resorting

See also: :class:`~k1lib.cli.utils.syncStepper`

:param timeline: sparser timeline
:param before: whether to use the timeline B's point before or after timeline A's point
:param startTok: token to use for A's samples where B has not gone into the picture yet
:param endTok: token to use for A's samples where B has yielded all elements"""  # latch
        self.timeline = timeline; self.before = before; self.startTok = startTok; self.endTok = endTok # latch
    def __ror__(self, it):                                                       # latch
        timeline = [[[-float("inf"), self.startTok]], self.timeline, [[float("inf"), self.endTok]]] | cli.joinSt() # latch
        before = self.before; lastT, lastT_desc = next(timeline); t, t_desc = next(timeline) # latch
        for row in it:                                                           # latch
            while row[0] > t:                                                    # latch
                lastT = t; lastT_desc = t_desc                                   # latch
                t, t_desc = next(timeline)                                       # latch
            yield [*row, lastT_desc if before else t_desc]                       # latch



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
For operations that feel like the termination of operations
"""
from collections import defaultdict
from typing import Iterator, Any
from k1lib.cli.init import BaseCli; import k1lib.cli.init as init
import numbers, numpy as np, k1lib, tempfile, os, sys, time, math, json, re
from k1lib import cli; from k1lib.cli.typehint import *
plt = k1lib.dep("matplotlib.pyplot")
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["stdout", "tee", "file", "pretty", "unpretty", "display", "headOut",
           "intercept", "plotImgs"]
settings = k1lib.settings.cli
class stdout(BaseCli):                                                           # stdout
    def __init__(self):                                                          # stdout
        """Prints out all lines. If not iterable, then print out the input
raw. Example::

    # prints out "0\\n1\\n2"
    range(3) | stdout()
    # same as above, but (maybe?) more familiar
    range(3) > stdout()

This is rarely used alone. It's more common to use :meth:`headOut`
for list of items, and :meth:`display` for tables."""                            # stdout
        super().__init__()                                                       # stdout
    def _typehint(self, inp): return None                                        # stdout
    def __ror__(self, it:Iterator[str]):                                         # stdout
        try:                                                                     # stdout
            it = iter(it)                                                        # stdout
            for line in it: print(line)                                          # stdout
        except TypeError: print(it)                                              # stdout
_defaultTeeF = lambda s: f"{s}\n"                                                # stdout
class tee(BaseCli):                                                              # tee
    def __init__(self, f=_defaultTeeF, s=None, every:int=1, delay:float=0):      # tee
        """Like the Linux ``tee`` command, this prints the elements to another
specified stream, while yielding the elements. Example::

    # prints "0) 0\\n1) 1\\n2) 2\\n3) 3\\n4) 4\\n" and returns [0, 1, 4, 9, 16]
    range(5) | tee() | apply(op() ** 2) | deref()

See also: :class:`~k1lib.cli.modifier.consume`

This cli is not exactly well-thoughout and is a little janky

:param f: element transform function. Defaults to just adding a new
    line at the end
:param s: stream to write to. Defaults to :attr:`sys.stdout`
:param every: only prints out 1 line in ``every`` lines, to limit print rate
:param delay: if subsequent prints are less than this number of seconds apart then don't print them""" # tee
        self.s = s or sys.stdout; self.f = f; self.every = every; self.delay = delay # tee
    def __ror__(self, it):                                                       # tee
        s = self.s; f = self.f; every = self.every; delay = self.delay           # tee
        lastTime = 0                                                             # tee
        for i, e in enumerate(it):                                               # tee
            if i % every == 0 and time.time()-lastTime > delay:                  # tee
                print(f"     \r{i}) {f(e)}", end="", file=s); lastTime = time.time() # tee
            yield e                                                              # tee
    def cr(self):                                                                # tee
        """Tee, but replaces the previous line. "cr" stands for carriage return.
Example::

    # prints "4" and returns [0, 1, 4, 9, 16]. Does print all the numbers in the middle, but is overriden
    range(5) | tee().cr() | apply(op() ** 2) | deref()"""                        # tee
        f = (lambda x: x) if self.f == _defaultTeeF else self.f                  # tee
        self.f = lambda s: f"{f(s)}"; return self                                # tee
    def crt(self):                                                               # tee
        """Like :meth:`tee.cr`, but includes an elapsed time text at the end.
Example::

    range(5) | tee().cr() | apply(op() ** 2) | deref()"""                        # tee
        beginTime = time.time()                                                  # tee
        f = (lambda x: x) if self.f == _defaultTeeF else self.f                  # tee
        self.f = lambda s: f"{f(s)}, {int(time.time() - beginTime)}s elapsed"; return self # tee
    def autoInc(self):                                                           # tee
        """Like :meth:`tee.crt`, but instead of printing the object, just print
the current index and time"""                                                    # tee
        beginTime = time.time(); autoInc = k1lib.AutoIncrement()                 # tee
        self.f = lambda s: f"{autoInc()}, {int(time.time()-beginTime)}s elapsed"; return self # tee
try:                                                                             # tee
    import PIL; hasPIL = True                                                    # tee
except: hasPIL = False                                                           # tee
class file(BaseCli):                                                             # file
    def __init__(self, fileName:str=None, flush:bool=False, mkdir:bool=False):   # file
        """Opens a new file for writing. This will iterate through
the iterator fed to it and put each element on a separate line. Example::

    # writes "0\\n1\\n2\\n" to file
    range(3) | file("test/f.txt")
    # same as above, but (maybe?) more familiar
    range(3) > file("text/f.txt")
    # returns ['0', '1', '2']
    cat("folder/f.txt") | deref()

If the input is a string, then it will just put the string into the
file and does not iterate through the string::

    # writes "some text\\n123" to file, default iterator mode like above
    ["some text", "123"] | file("test/f.txt")
    # same as above, but this is a special case when it detects you're piping in a string
    "some text\\n123" | file("test/f.txt")

If the input is a :class:`bytes` object or an iterator of :class:`bytes`, then it
will open the file in binary mode and dumps the bytes in::

    # writes bytes to file
    b'5643' | file("test/a.bin")
    [b'56', b'43'] >> file("test/a.bin")
    # returns ['56435643']
    cat("test/a.bin") | deref()

If the input is a :class:`PIL.Image.Image` object, then it will just save the image in
the file::

    # creates an random image and saves it to a file
    torch.randn(100, 200) | toImg() | file("a.png")

Reminder that the image pixel range is expected to be from 0 to 255. You
can create temporary files on the fly by not specifying a file name::

    # creates temporary file
    url = range(3) > file()
    # returns ['0', '1', '2']
    cat(url) | deref()

This can be especially useful when integrating with shell scripts that wants to
read in a file::

    seq1 = "CCAAACCCCCCCTCCCCCGCTTC"
    seq2 = "CCAAACCCCCCCCTCCCCCCGCTTC"
    # use "needle" program to locally align 2 sequences
    None | cmd(f"needle {seq1 > file()} {seq2 > file()} -filter")

You can also append to file with the ">>" operator::

    url = range(3) > file()
    # appended to file
    range(10, 13) >> file(url)
    # returns ['0', '1', '2', '10', '11', '12']
    cat(url) | deref()

:param fileName: if not specified, create new temporary file and returns the url
    when pipes into it
:param flush: whether to flush to file immediately after every iteration
:param mkdir: whether to recursively make directories going to the file location or not""" # file
        super().__init__(); self.fileName = fileName; self.flush = flush; self.mkdir = mkdir # file
        self.append = False # whether to append to file rather than erasing it   # file
    def __ror__(self, it:Iterator[str]) -> None:                                 # file
        super().__ror__(it); fileName = self.fileName; flushF = (lambda f: f.flush()) if self.flush else (lambda _: 0) # file
        if fileName is None:                                                     # file
            f = tempfile.NamedTemporaryFile()                                    # file
            fileName = f.name; f.close()                                         # file
        fileName = os.path.expanduser(fileName); firstLine = None                # file
        if self.mkdir: os.makedirs(os.path.dirname(fileName), exist_ok=True)     # file
        if hasPIL and isinstance(it, PIL.Image.Image): it.save(fileName); return fileName # file
        if isinstance(it, str): it = [it]; text = True                           # file
        elif isinstance(it, bytes): text = False                                 # file
        else:                                                                    # file
            it = iter(it); sentinel = object(); firstLine = next(it, sentinel)   # file
            if firstLine is sentinel: # no elements at all                       # file
                with open(fileName, "w") as f: f.write("")                       # file
                return fileName                                                  # file
            text = not isinstance(firstLine, bytes)                              # file
        if text:                                                                 # file
            with open(fileName, "a" if self.append else "w") as f:               # file
                if firstLine is not None: f.write(f"{firstLine}\n")              # file
                for line in it: f.write(f"{line}\n"); flushF(f)                  # file
        else:                                                                    # file
            with open(fileName, "ab" if self.append else "wb") as f:             # file
                if firstLine is not None:                                        # file
                    f.write(firstLine)                                           # file
                    for e in it: f.write(e); flushF(f)                           # file
                else: f.write(it)                                                # file
        return fileName                                                          # file
    def __rrshift__(self, it):                                                   # file
        self.append = True # why do this? because `a | b >> c` will be interpreted as `a | (b >> c)` # file
        if isinstance(it, BaseCli): return cli.serial(it, self)                  # file
        else: return self.__ror__(it)                                            # file
    @property                                                                    # file
    def name(self):                                                              # file
        """File name of this :class:`file`"""                                    # file
        return self.fileName                                                     # file
class pretty(BaseCli):                                                           # pretty
    def __init__(self, delim="", left=True):                                     # pretty
        """Pretty-formats a table, or a list of tables.
Example::

    # These 2 statements are pretty much the same
    [range(10), range(10, 20)] | head(5) | pretty() > stdout()
    [range(10), range(10, 20)] | display()

They both print::

    0    1    2    3    4    5    6    7    8    9
    10   11   12   13   14   15   16   17   18   19

This can also pretty-formats multiple tables::

    [[range(10), range(10, 20)], [["abc", "defff"], ["1", "1234567"]]] | ~pretty() | joinStreams() | stdout()

This will print::

    0     1         2    3    4    5    6    7    8    9
    10    11        12   13   14   15   16   17   18   19
    abc   defff
    1     1234567

:param delim: delimiter between elements within a row. You might want
    to set it to "|" to create an artificial border or sth
:param left: whether to left or right-align each element"""                      # pretty
        self.delim = delim; self.inverted = False; self.left = left              # pretty
    def _typehint(self, inp): return tIter(str)                                  # pretty
    def __ror__(self, it) -> Iterator[str]:                                      # pretty
        inv = self.inverted; delim = self.delim; left = self.left                # pretty
        if inv: tables = [[list(i1) for i1 in i2] for i2 in it]                  # pretty
        else: tables = [[list(i1) for i1 in it]]                                 # pretty
        widths = defaultdict(lambda: 0)                                          # pretty
        for table in tables:                                                     # pretty
            for row in table:                                                    # pretty
                for i, e in enumerate(row):                                      # pretty
                    e = f"{e}"; row[i] = e                                       # pretty
                    widths[i] = max(len(e), widths[i])                           # pretty
        def gen(table):                                                          # pretty
            if left:                                                             # pretty
                for row in table: yield delim.join(e.rstrip(" ").ljust(w+3) for w, e in zip(widths.values(), row)) # pretty
            else:                                                                # pretty
                for row in table: yield delim.join(e.rstrip(" ").rjust(w+3) for w, e in zip(widths.values(), row)) # pretty
        if inv: return tables | cli.apply(gen)                                   # pretty
        else: return gen(tables[0])                                              # pretty
    def __invert__(self): self.inverted = not self.inverted; return self         # pretty
    def _jsF(self, meta):                                                        # pretty
        fIdx = init._jsFAuto(); dataIdx = init._jsDAuto()                        # pretty
        return f"const {fIdx} = ({dataIdx}) => {dataIdx}.pretty({json.dumps(self.delim)}, {cli.kjs.v(self.inverted)})", fIdx # pretty
def display(lines:int=10):                                                       # display
    """Convenience method for displaying a table.
Pretty much equivalent to ``head() | pretty() | stdout()``.

See also: :class:`pretty`"""                                                     # display
    f = pretty() | stdout()                                                      # display
    if lines is None: return f                                                   # display
    else: return cli.head(lines) | f                                             # display
def headOut(lines:int=10):                                                       # headOut
    """Convenience method for head() | stdout()"""                               # headOut
    if lines is None: return stdout()                                            # headOut
    else: return cli.head(lines) | stdout()                                      # headOut
class unpretty(BaseCli):                                                         # unpretty
    def __init__(self, ncols:int=None, left=True, headers=None):                 # unpretty
        """Takes in a stream of strings, assumes it's a table, and tries to
split every line into multiple columns. Example::

    # returns ['0   1   2   ', '3   4   5   ', '6   7   8   ']
    a = range(10) | batched(3) | pretty() | deref()
    # returns [['0   ', '1   ', '2   '], ['3   ', '4   ', '5   '], ['6   ', '7   ', '8   ']]
    a | unpretty(3) | deref()

This cli will take the number of columns requested and try to split into a table by analyzing
at what character column does it transition from a space to a non-space (left align), or from
a non-space to a space (right align). Then the first ``ncols`` most popular transitions are
selected.

Sometimes this is not robust enough, may be some of your columns have lots of empty elements,
then the transition counts will be skewed, making it split up at strange places. In those cases,
you can specify the headers directly, like this::

    # returns ['a   b   c    ', '3   5   11   ', '4   6   7    ']
    a = [["a", 3, 4], ["b", 5, 6], ["c", 11, 7]] | transpose() | pretty() | deref()
    # returns [['a   ', 'b   ', 'c    '], ['3   ', '5   ', '11   '], ['4   ', '6   ', '7    ']]
    a | unpretty(headers=["a", "b", "c"]) | deref()

:param ncols: number of columns
:param left: whether the data is left or right aligned
:param header:"""                                                                # unpretty
        self.ncols = ncols; self.left = left; self.headers = headers             # unpretty
        self.pat = re.compile(" [^ ]+") if left else re.compile("[^ ]+ ")        # unpretty
    def __ror__(self, it):                                                       # unpretty
        ncols = self.ncols; left = self.left; pat = self.pat; headers = self.headers # unpretty
        if headers is not None: ncols = len(headers)                             # unpretty
        if ncols < 1: raise Exception(f"Does not make sense to unpretty() into {ncols} columns") # unpretty
        if ncols == 1: return it                                                 # unpretty
        if headers is None:                                                      # unpretty
            try: len(it)                                                         # unpretty
            except: it = list(it)                                                # unpretty
            splits = it | cli.head(10000) | (cli.apply(lambda x: (m.start()+1 for m in re.finditer(pat, x))) if left else cli.apply(lambda x: (m.end()-1 for m in re.finditer(pat, x))))\
                | cli.joinSt() | cli.count() | ~cli.sort() | cli.cut(1) | cli.head(ncols-1) | cli.sort(None) | cli.aS(list) # unpretty
        else:                                                                    # unpretty
            firstRow, it = it | cli.peek()                                       # unpretty
            if it == []: return []                                               # unpretty
            splits = sorted([firstRow.find(h) for h in headers])[1:]             # unpretty
        if ncols == 2: c = splits[0]; return ([row[:c],row[c:]] for row in it)   # unpretty
        a,*r,b = splits; s = splits | cli.window(2) | ~cli.apply(lambda x,y: f"x[{x}:{y}], ") | cli.join("") # unpretty
        f = eval(f"lambda x: [x[:{a}], {s}x[{b}:]]"); return (f(row) for row in it) # unpretty
def tab(text, pad="    "):                                                       # tab
    return "\n".join([pad + line for line in text.split("\n")])                  # tab
class intercept(BaseCli):                                                        # intercept
    def __init__(self, f=None, raiseError:bool=True):                            # intercept
        """Intercept flow at a particular point, analyze the object piped in, and
raises error to stop flow. Example::

    3 | intercept()

:param f: prints out the object transformed by this function
:param raiseError: whether to raise error when executed or not."""               # intercept
        self.f = f or cli.shape(); self.raiseError = raiseError                  # intercept
    def __ror__(self, s):                                                        # intercept
        print(type(s)); print(self.f(s))                                         # intercept
        if self.raiseError: raise RuntimeError("intercepted")                    # intercept
        return s                                                                 # intercept
class plotImgs(BaseCli):                                                         # plotImgs
    def __init__(self, col=5, aspect=1, fac=2, axis=False, table=False, im=False): # plotImgs
        """Plots a bunch of images at the same time in a table.
Example::

    # plots all images
    [torch.randn(10, 20), torch.randn(20, 10)] | plotImgs()
    # plots all images with titles
    [[torch.randn(10, 20), "img 1"], [torch.randn(20, 10), "img 2"]] | plotImgs()

If you have multiple rows with different number of images, you can
plot that with this too, just set ``table=True`` like this::

    [[torch.randn(10, 20), torch.randn(20, 10)], [torch.randn(10, 20)]] | plotImgs(table=True)

There's another cli that kinda does what this does: :class:`~k1lib.cli.utils.sketch`. You have
more control over there, and it does roughly what this cli does, but the typical usage is
different. This is more for plotting static, throwaway list of 2d arrays, like training set
images, where as :class:`~k1lib.cli.utils.sketch` is more about plotting results of detailed
analyses.

:param col: number of columns in the table. If explicitly None, it will turn
    into the number of images fed. Not available if ``table=True``
:param aspect: aspect ratio of each images, or ratio between width and height
:param fac: figsize factor. The higher, the more resolution
:param axis: whether to display the axis or not
:param table: whether to plot using table mode
:param im: if True, returns an image"""                                          # plotImgs
        self.col = col; self.fac = fac; self.axis = axis; self.aspect = aspect; self.table = table; self.im = im # plotImgs
    def __ror__(self, imgs):                                                     # plotImgs
        imgs = imgs | cli.deref(); col = self.col; fac = self.fac; aspect = self.aspect**0.5 # plotImgs
        if not self.table: # main code                                           # plotImgs
            if len(imgs) == 0: return                                            # plotImgs
            if col is None or col > len(imgs): col = len(imgs)                   # plotImgs
            n = math.ceil(len(imgs)/col)                                         # plotImgs
            fig, axes = plt.subplots(n, col, figsize=(col*fac*aspect, n*fac/aspect)); # plotImgs
            axes = axes.flatten() if isinstance(axes, np.ndarray) else [axes]    # plotImgs
            for ax, im in zip(axes, imgs):                                       # plotImgs
                plt.sca(ax)                                                      # plotImgs
                if isinstance(im, (list, tuple)): plt.imshow(im[0]); plt.title(im[1]) # plotImgs
                else: plt.imshow(im)                                             # plotImgs
                if not self.axis: ax.axis("off")                                 # plotImgs
            for i in range(len(imgs), len(axes)): axes[i].remove() # removing leftover axes # plotImgs
        else:                                                                    # plotImgs
            if col != 5: raise Exception("Currently in table mode, can't set `col` parameter") # change this value to match col's default value # plotImgs
            h = imgs | cli.shape(0); w = imgs | cli.shape(0).all() | cli.toMax() # plotImgs
            fig, axes = plt.subplots(h, w, figsize=(w*fac*aspect, h*fac/aspect)); # plotImgs
            for rAx, rIm in zip(axes, imgs):                                     # plotImgs
                for cAx, cIm in zip(rAx, rIm):                                   # plotImgs
                    plt.sca(cAx)                                                 # plotImgs
                    if isinstance(cIm, (list, tuple)): plt.imshow(cIm[0]); plt.title(cIm[1]) # plotImgs
                    else: plt.imshow(cIm)                                        # plotImgs
                    if not self.axis: cAx.axis("off")                            # plotImgs
                for i in range(len(rIm), len(rAx)): rAx[i].remove() # removing leftover axes # plotImgs
        plt.tight_layout()                                                       # plotImgs
        if self.im: return plt.gcf() | cli.toImg()                               # plotImgs



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This includes helper clis that make it quick to graph graphviz plots."""
__all__ = ["markers", "Markers", "map3d"]
import re, k1lib, math, os, numpy as np, io, json, base64, html
from k1lib.cli.init import BaseCli; import k1lib.cli as cli, k1lib.cli.init as init
cv2 = k1lib.dep("cv2")
aruco = k1lib.dep("cv2.aruco")
from collections import deque, defaultdict
settings = k1lib.settings.cli
class markers(BaseCli):                                                          # markers
    def __init__(self, d:str=None):                                              # markers
        """Detect aruco markers. Returns :class:`Markers`.
Example::

    im = "some/image.jpg" | toImg()
    markers = im | kcv.markers()

    markers # run this in a notebook cell to quickly view where the bounding boxes are
    markers | item() # grab first detection, a tuple (aruco dictionary, coords (with shape (4,2)), number)
    markers[0] # same as above
    markers | deref() # dereference into just lists, instead of kcv.Markers object

For a moderately complex scene with 20 markers of size (1000, 800), this should take ~15ms to run. The same
scene of size (4000, 3000) should take ~170ms to run

:param d: aruco dictionary string, like 'DICT_7X7_1000'. If specified, then will only look in
    those dictionaries. If not specified will look in all dictionaries"""        # markers
        self.ds = [d] if d else ["DICT_4X4_1000", "DICT_5X5_1000", "DICT_6X6_1000", "DICT_7X7_1000"] # markers
    def __ror__(self, im):                                                       # markers
        arr = im | cli.toNdArray(np.uint8) | cli.op().transpose(1, 2, 0)         # markers
        return Markers(im, self.ds | cli.apply(lambda d: [d, cv2.aruco.detectMarkers(arr, aruco.Dictionary_get(getattr(aruco, d)))[:2]]) | cli.deref() | ~cli.filt("x[1] is None", 1) | cli.apply(cli.item().all(2) | cli.T() , 1) | cli.ungroup(False) | cli.deref()) # markers
class Markers:                                                                   # Markers
    def __init__(self, im, ms):                                                  # Markers
        """Resulting Markers object, obtained from :class:`markers` cli, not
intended to be instantiated by the end user."""                                  # Markers
        self.im = im; self._ms = ms                                              # Markers
    def __getitem__(self, s): return self._ms[s]                                 # Markers
    def __iter__(self): return iter(self._ms)                                    # Markers
    def __len__(self): return len(self._ms)                                      # Markers
    def __repr__(self): return f"<Markers n={len(self)} im.shape={self.im | cli.shape() | cli.op()[::-1]} ds={self._ms | cli.cut(0) | cli.count() | ~cli.sort() | cli.cut(1) | cli.deref()}>" # Markers
    def _repr_html_(self):                                                       # Markers
        im = self.im; ms = self._ms; p5 = k1lib.p5; p5.newSketch(*im | cli.shape(), False) # Markers
        p5.image(im, 0, 0); p5.background(200, 220)                              # Markers
        for _dict, coords, num in ms:                                            # Markers
            a,b,c,d = coords                                                     # Markers
            p5.stroke(255, 0, 0);   p5.line(*a,*b)                               # Markers
            p5.stroke(0, 255, 0);   p5.line(*b,*c)                               # Markers
            p5.stroke(0, 0, 255);   p5.line(*c,*d)                               # Markers
            p5.stroke(255, 255, 0); p5.line(*d,*a)                               # Markers
            p5.stroke(255, 0, 255); p5.text(f"{num}", *a)                        # Markers
        return f"{html.escape(self.__repr__())}<br>{p5.svg()}"                   # Markers
def map3d(ws,cs):                                                                # map3d
    """Creates 2 functions that maps from world points (ws) to camera points (cs).
Example::

    ws = np.random.randn(10, 3)
    cs = np.copy(ws); cs[:,1] = -cs[:,1]; cs[:,2] = -cs[:,2]

    f1, f2 = kcv.map3d(ws, cs)

    f1(ws) # returns matrix similar to cs
    f2(cs) # returns matrix similar to ws
"""                                                                              # map3d
    wc = np.mean(ws, axis=0); wsp = ws - wc                                      # map3d
    cc = np.mean(cs, axis=0); csp = cs - cc                                      # map3d
    H = wsp.T @ csp; U, S, Vt = np.linalg.svd(H)                                 # map3d
    R = np.dot(U, Vt); R_inv = np.linalg.inv(R)                                  # map3d
    t = cc - np.dot(R, wc)                                                       # map3d
    # main equation is given by chatgpt, but the transformation for some reason is translated by a # map3d
    # fixed vector, so this is to figure out the correct bias vector to add to   # map3d
    b1 = cc-(ws@R+t) | cli.T() | cli.toMean().all()                              # map3d
    b2 = ws-(cs-t)@R_inv | cli.T() | cli.toMean().all()                          # map3d
    return lambda vs: (vs)@R+t+b1, lambda vs: (vs-t)@R_inv+b2                    # map3d



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import re, difflib, k1lib
from typing import Dict, Union, List, Optional
__all__ = ["Eqn", "Eqns", "System"]
settings = k1lib.Settings()
settings.spaceBetweenValueSymbol = True
settings.eqnPrintExtras = True
k1lib.settings.add("eqn", settings, "from k1lib.eqn module");
class Eqn:                                                                       # Eqn
    def __init__(self, system:"System"):                                         # Eqn
        """Creates a blank equation. Not expected to be instantiated by
the end user."""                                                                 # Eqn
        self.system = system                                                     # Eqn
        self.terms:Dict[str, float] = {}                                         # Eqn
    def removeZeros(self):                                                       # Eqn
        self.terms = {k: v for k, v in self.terms.items() if abs(v) > 1e-6}; return self # Eqn
    def parse(self, line:str):                                                   # Eqn
        line = line.strip().replace(" +", "+").replace("+ ", "+").replace(" ->", "->").replace("-> ", "->") # Eqn
        reactants, products = line.split("->")                                   # Eqn
        for i, side in enumerate(line.split("->")):                              # Eqn
            sign = i * 2 - 1                                                     # Eqn
            for e in side.split("+"): # side is reactants or products            # Eqn
                e = e.strip()                                                    # Eqn
                number = re.findall("^[0-9.\/]*", e)[0]                          # Eqn
                term, number = (e, 1) if number == "" else (e[e.find(number) + len(number):], eval(str(number))) # Eqn
                term = term.strip(); self.system.terms.add(term)                 # Eqn
                if term not in self: self[term] = 0                              # Eqn
                self[term] += sign * number                                      # Eqn
        return self.removeZeros()                                                # Eqn
    def save(self):                                                              # Eqn
        """Saves this (potentially new) equation to the system, so that it
can be used directly later on"""                                                 # Eqn
        self.system.parse(str(self)); return self                                # Eqn
    def __contains__(self, x:str):                                               # Eqn
        """Whether a term is in this equation"""                                 # Eqn
        return x in self.terms                                                   # Eqn
    def __getattr__(self, term:str):                                             # Eqn
        """Gets the value of the term in this equation. Negative if on
consumer side, positive if on producer side"""                                   # Eqn
        if term in self.terms: return self.terms[term]                           # Eqn
        else: return 0                                                           # Eqn
    def __setitem__(self, idx:str, value:float): self.terms[idx] = value; return self # Eqn
    def __getitem__(self, idx:str):                                              # Eqn
        """Same as :meth:`__getattr__`"""                                        # Eqn
        return getattr(self, idx)                                                # Eqn
    def __iter__(self):                                                          # Eqn
        """Yields key:value pairs"""                                             # Eqn
        for k, v in self.terms.items(): yield k, v                               # Eqn
    def __len__(self):                                                           # Eqn
        """Returns number of terms in this equation"""                           # Eqn
        return len(self.terms)                                                   # Eqn
    def __hash__(self): return hash(tuple(self.terms.keys()))                    # Eqn
    def __str__(self):                                                           # Eqn
        a = " + ".join((f"{-v}{k}" for k, v in self.terms.items() if v < 0))     # Eqn
        b = " + ".join((f"{v}{k}" for k, v in self.terms.items() if v > 0))      # Eqn
        return f"{a} -> {b}"                                                     # Eqn
    def copy(self):                                                              # Eqn
        answer = Eqn(self.system)                                                # Eqn
        answer.terms = dict(self.terms); return answer                           # Eqn
    def __repr__(self, printExtras=None):                                        # Eqn
        space = " " if settings.spaceBetweenValueSymbol else ""                  # Eqn
        def formatValue(value:float):                                            # Eqn
            if abs(value - 1) < 1e-9: return ""                                  # Eqn
            if abs(value - round(value)) < 1e-9:                                 # Eqn
                return f"{round(value)}{space}"                                  # Eqn
            return f"{round(value, 3)}{space}"                                   # Eqn
        a = " + ".join((f"{formatValue(-v)}{k}" for k, v in self.terms.items() if v < 0)) # Eqn
        b = " + ".join((f"{formatValue(v)}{k}" for k, v in self.terms.items() if v > 0)) # Eqn
        answer = f"{a} \033[1m->\033[0m {b}"                                     # Eqn
        printExtras = printExtras if printExtras is not None else settings.eqnPrintExtras # Eqn
        return answer if not printExtras else f"""{answer}. Can...
- "MJ" in eqn: to check whether this equation has a specific term
- eqn["MJ"], or eqn.MJ: to get the actual value of the term
- eqn.terms: to get dict of all term -> values
- eqn["MJ"] = 5: to modify a term's value
- eqn * 2: to use normal math operations on the entire equation
- eqn1 @ eqn2: to try to zero out some common terms, useful for unit conversions
- eqn1 == eqn2: see if 2 equations are the same, scale invariant
- for term, value in eqn: to loop over every term and its value
- len(eqn): to get number of terms in the equation
- eqn.copy()"""                                                                  # Eqn
    def __mul__(self, number:float):                                             # Eqn
        answer = self.copy()                                                     # Eqn
        answer.terms = {k: v*number for k, v in self.terms.items()}              # Eqn
        return answer.removeZeros()                                              # Eqn
    def __rmul__(self, number:float): return self.__mul__(number)                # Eqn
    def __neg__(self): return -1 * self                                          # Eqn
    def __truediv__(self, number:float):                                         # Eqn
        answer = self.copy()                                                     # Eqn
        answer.terms = {k: v/number for k, v in self.terms.items()}              # Eqn
        return answer.removeZeros()                                              # Eqn
    def __rtruediv__(self, number:float): raise Exception("Can't be divided by a number. It doesn't mean anything") # Eqn
    def __add__(self, eqn):                                                      # Eqn
        answer = self.copy(); answer.terms = {}                                  # Eqn
        for term, value in self: answer[term] = value + eqn[term]                # Eqn
        for term, value in eqn:                                                  # Eqn
            if term not in answer: answer[term] = value + self[term]             # Eqn
        return answer.removeZeros()                                              # Eqn
    def __sub__(self, eqn): return self + -1*eqn                                 # Eqn
    def __eq__(self, eqn):                                                       # Eqn
        if len(self) != len(eqn): return False                                   # Eqn
        if set(self.terms.keys()) != set(eqn.terms.keys()): return False         # Eqn
        term = list(self.terms.keys())[0]                                        # Eqn
        eqn = eqn * self[term] / eqn[term]                                       # Eqn
        for term, value in self:                                                 # Eqn
            if abs(self[term] - eqn[term]) > 1e-9: return False                  # Eqn
        return True                                                              # Eqn
    def sharedTerms(self, eqn:"Eqn") -> List[str]:                               # Eqn
        """Gets a list of shared terms between this equation and the
specified one."""                                                                # Eqn
        ts = set(self.terms.keys())                                              # Eqn
        return [t for t in eqn.terms.keys() if t in ts]                          # Eqn
    def join(self, eqn:"Eqn", term:str) -> "Eqn":                                # Eqn
        """Tries to cancel out this equation with another equation at the
specified term. Example::

    s = eqn.System(\"\"\"a + b -> c + d
    c + 2e -> f\"\"\")
    s.a.c.join(s.c.f, "c") # returns the equation "a + b + 2e -> d + f"

For simpler cases, where the shared term to be joined is obvious, use
:meth:`__matmul__` instead"""                                                    # Eqn
        return self + eqn * (-self[term]/eqn[term])                              # Eqn
    def __matmul__(self, eqn:"Eqn") -> "Eqn":                                    # Eqn
        """Convenience method that does the same thing as :meth:`join`.
Example::

    s = eqn.System(\"\"\"a + b -> c + d
    c + 2e -> f\"\"\")
    s.a.c @ s.c.f # returns the equation "a + b + 2e -> d + f"

Preference order of which term to join:

1) If term is on producer side of ``self``, and consumer side of ``eqn``
2) If term is on consumer side of ``self``, and producer side of ``eqn``
3) Other cases"""                                                                # Eqn
        sharedTerms = self.sharedTerms(eqn)                                      # Eqn
        def sortF(term):                                                         # Eqn
            if self[term] > 0 and eqn[term] < 0: return 0                        # Eqn
            if self[term] < 0 and eqn[term] > 0: return 1                        # Eqn
            return 2                                                             # Eqn
        sharedTerms = sorted(sharedTerms, key=sortF)                             # Eqn
        if len(sharedTerms) == 0: return None                                    # Eqn
        return self.join(eqn, sharedTerms[0])                                    # Eqn
    def round(self, term:str, amount:float=10) -> "Eqn":                         # Eqn
        """Rounds the equation off, so that the term's value
is the specified amount. For aesthetic purposes mainly. Example::

    s = eqn.System("a + b -> 2c")
    s.a.c.round("c", 5) # returns the equation "2.5a + 2.5b -> 5c"'"""           # Eqn
        if term not in self: raise AttributeError(term)                          # Eqn
        return self * amount / self[term]                                        # Eqn
    def __round__(self, term:str=None) -> "Eqn":                                 # Eqn
        """Like :meth:`round`, but more Pythonic?

:param term: Can be any of these:

    - None
    - str
    - Union[int, float]
    - Tuple[str, float]"""                                                       # Eqn
        defaultTerm = list(self.terms.keys())[-1]                                # Eqn
        if term is None: return self.round(defaultTerm, 1)                       # Eqn
        elif isinstance(term, (tuple, list)): return self.round(*term)           # Eqn
        elif isinstance(term, str): return self.round(term, 1)                   # Eqn
        elif k1lib.isNumeric(term): return self.round(defaultTerm, term)         # Eqn
        else: raise AttributeError(f"Don't understand {term}")                   # Eqn
class Eqns:                                                                      # Eqns
    def __init__(self, system:"System", eqns:List[Eqn], focusTerm:str=None):     # Eqns
        """Creates a new list of equations. Not expected to be instantiated
by the end user.

:param system: injected :class:`System`
:param eqns: list of equations
:param focusTerm: if the list of equations are from the result of focusing
    in a single term, then use this parameter to prioritize certain search
    parameters.
"""                                                                              # Eqns
        self.system = system; self.eqns = eqns; self.terms = set()               # Eqns
        for eqn in eqns: self.terms.update(eqn.terms.keys())                     # Eqns
        self.focusTerm = focusTerm                                               # Eqns
    def __getitem__(self, idx:Union[int, str]) -> Optional[Eqn]:                 # Eqns
        """If int, return the equation with that index. Not really helpful
for exploring the system of equations, but good for automated scripts

If string, then effectively the same as :meth:`__getattr__`
"""                                                                              # Eqns
        return self.eqns[idx] if isinstance(idx, int) else getattr(self, idx)    # Eqns
    def __getattr__(self, term:str) -> Optional[Eqn]:                            # Eqns
        """Picks out a specific :class:`Eqn` that has the specified term.
Prefer shorter equations, and the returned :class:`Eqn` always have the
term on the products side. Meaning::

    eqns = eqn.System("a + 2b -> c").b # gets an Eqns object with that single equation
    eqns.a # gets the equation "c -> a + 2b" instead

This is a convenience way to search for equations. If you need more
granularity, use :meth:`pick` instead"""                                         # Eqns
        chosenEqns = []                                                          # Eqns
        for eqn in self.eqns:                                                    # Eqns
            if term in eqn:                                                      # Eqns
                chosenEqns.append(eqn if eqn[term] > 0 else -eqn)                # Eqns
        chosenEqns = sorted(chosenEqns, key=lambda eqn: len(eqn))                # Eqns
        return None if len(chosenEqns) == 0 else chosenEqns[0]                   # Eqns
    def pick(self, *terms:List[str]) -> Optional[Eqn]:                           # Eqns
        """Like the quick method (:meth:`__getattr__`), but here, picks
equations more carefully, with selection for multiple terms. Example::

    s = eqn.System(\"\"\"a + 2b -> c
    b + c -> d
    a -> 3d
    a + b + c -> 2d\"\"\")
    s.a.pick("b", "d") # returns last equation

As you can see, it's impossible to pick out the last equation using
:meth:`__getattr__` alone, as they will all prefer the shorter equations,
so this is where :meth:`pick` can be useful."""                                  # Eqns
        chosenEqns = []; t = self.focusTerm or terms[0]                          # Eqns
        for eqn in self.eqns:                                                    # Eqns
            if all((term in eqn for term in terms)):                             # Eqns
                chosenEqns.append(eqn if eqn[t] > 0 else -eqn)                   # Eqns
        chosenEqns = sorted(chosenEqns, key=lambda eqn: len(eqn))                # Eqns
        return None if len(chosenEqns) == 0 else chosenEqns[0]                   # Eqns
    def __dir__(self):                                                           # Eqns
        """Returns the list of terms in every equation here. Useful for
tab completion."""                                                               # Eqns
        return list(self.terms)                                                  # Eqns
    def __repr__(self):                                                          # Eqns
        end = """Can...
- eqns[i]: to get the 'i'th equation
- eqns.C: to pick out the first equation that has term 'C'"""                    # Eqns
        if self.focusTerm == None:                                               # Eqns
            eqns = "\n".join([f"{i}. {eqn.__repr__(printExtras=False)}" for i, eqn in enumerate(self.eqns)]) # Eqns
            return f"""Equations:\n{eqns}\n\n{end}"""                            # Eqns
        else:                                                                    # Eqns
            consumingEqns = []; producingEqns = []                               # Eqns
            for eqn in self.eqns:                                                # Eqns
                if eqn[self.focusTerm] < 0:                                      # Eqns
                    consumingEqns.append(f"{eqn.__repr__(printExtras=False)}")   # Eqns
                else: producingEqns.append(f"{eqn.__repr__(printExtras=False)}") # Eqns
            consumingEqns = "\n".join([f"{i}. {eqn}" for i, eqn in enumerate(consumingEqns)]) # Eqns
            producingEqns = "\n".join([f"{i}. {eqn}" for i, eqn in enumerate(producingEqns)]) # Eqns
            return f"""Consumers:\n{consumingEqns}\n\nProducers:\n{producingEqns}\n\n{end}""" # Eqns
class System:                                                                    # System
    def __init__(self, strToParse:str=None):                                     # System
        """Creates a new system of equations.

:param strToParse: if specified, then it gets feed into :meth:`parse`"""         # System
        self.terms = set()                                                       # System
        self.eqns = []                                                           # System
        if strToParse is not None: self.parse(strToParse)                        # System
    def parse(self, lines:str) -> "System":                                      # System
        """Parses extra equations and saves them to this :class:`System`"""      # System
        lines = (line for line in lines.split("\n") if line != "" and not line.startswith("#")) # System
        self.eqns += [Eqn(self).parse(line) for line in lines if not line.startswith("#")] # System
        self.eqns = list(set(self.eqns))                                         # System
        return self                                                              # System
    def spellCheck(self):                                                        # System
        """Runs a spell check to find out terms that are pretty similar to
each other"""                                                                    # System
        print("Similar terms:"); terms = list(self.terms)                        # System
        for i, iTerm in enumerate(terms):                                        # System
            for j, jTerm in enumerate(terms[i+1:]):                              # System
                if iTerm[:-1] == jTerm[:-1]: continue                            # System
                if abs(len(iTerm) - len(jTerm)) > 2: continue                    # System
                r = difflib.SequenceMatcher(None, iTerm, jTerm).ratio()          # System
                if r < 0.9: continue                                             # System
                print(f"- {round(r*100)}% similar: {iTerm}, {jTerm}")            # System
    def __len__(self): return len(self.eqns)                                     # System
    def __getitem__(self, idx:int) -> Eqn:                                       # System
        """Picks out the i'th equation from the list of equations. Useful
for automated scripts"""                                                         # System
        return self.eqns[idx]                                                    # System
    def __getattr__(self, term:str) -> Eqns:                                     # System
        """Picks out equations that has the term"""                              # System
        return Eqns(self, [eqn for i, eqn in enumerate(self.eqns) if term in eqn], focusTerm=term) # System
    def __dir__(self):                                                           # System
        """Returns the list of terms in every equation here. Useful for
tab completion."""                                                               # System
        return list(self.terms)                                                  # System
    def __repr__(self):                                                          # System
        return f"""System of {len(self)} equations:\n{Eqns(self, self.eqns)}\n
Can...
- s[i]: to get a specific equation
- s.C: to get all equations that involve a specific substance "C"
- s.spellCheck(): to check if there are terms that are close to each other
"""                                                                              # System
